#! python3

"""
https://www.manhuaren.com/manhua-mofashaonvshenmedeyijinggoulela/
"""

import json
import re
from html import unescape
from urllib.parse import urljoin

from deno_vm import eval

from ..core import Episode

domain = ["www.manhuaren.com"]
name = "漫畫人"

def get_title(html, url):
	name = re.search('DM5_COMIC_MNAME=("[^"]+")', html).group(1)
	return json.loads(name)

def get_episodes(html, url):
	s = []
	pattern = 'href="(/m\d+/)[^>]+chapteritem[^>]+>([^<]+)'
	for match in re.finditer(pattern, html):
		ep_url, title = match.groups()
		s.append(Episode(unescape(title), urljoin(url, ep_url)))
	return s[::-1]
	
def get_images(html, url):
	js = re.search(r"(eval\([\s\S]+?)</script", html).group(1)
	return eval(js + ";newImgs")



#! python3

import re
import time
import imghdr
from contextlib import contextmanager
from pprint import pformat
from threading import Lock
from urllib.parse import quote, urlsplit, urlunsplit
from mimetypes import guess_extension

import enlighten
import requests
import puremagic
from worker import async_, await_, sleep, Defer

from .config import setting
from .io import content_write
from .profile import get as profile

default_header = {
	"User-Agent": "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/67.0.3396.99 Safari/537.36",
	"Accept-Language": "zh-tw,zh;q=0.8,en-us;q=0.5,en;q=0.3",
	"Accept-Encoding": "gzip, deflate"
}

cooldown = {}
grabber_pool = {}
grabber_pool_lock = Lock()
pb_manager = enlighten.get_manager()

@contextmanager
def get_request_lock(url):
	domain = urlsplit(url).hostname
	defer = Defer()
	try:
		with grabber_pool_lock:
			last_defer = grabber_pool.get(domain)
			grabber_pool[domain] = defer
		if last_defer:
			last_defer.get()
		yield
	finally:
		@async_
		def _():
			time.sleep(cooldown.get(domain, 0))
			defer.resolve(None)

def quote_unicode(s):
	"""Quote unicode characters only."""
	return quote(s, safe=r"/ !\"#$%&'()*+,:;<=>?@[\\]^`{|}~")

def quote_loosely(s):
	"""Quote space and others in path part.

	Reference:
	  http://stackoverflow.com/questions/120951/how-can-i-normalize-a-url-in-python
	"""
	return quote(s, safe="%/:=&?~#+!$,;'@()*[]")

def safeurl(url):
	"""Return a safe url, quote the unicode characters.

	This function should follow this rule:
	  safeurl(safeurl(url)) == safe(url)
	"""
	scheme, netloc, path, query, _fragment = urlsplit(url)
	return urlunsplit((scheme, netloc, quote_loosely(path), query, ""))

def quote_unicode_dict(d):
	"""Return a safe header, quote the unicode characters."""
	for key, value in d.items():
		d[key] = quote_unicode(value)
	
def grabber_log(*args):
	if setting.getboolean("errorlog"):
		content = time.strftime("%Y-%m-%dT%H:%M:%S%z") + "\n" + pformat(args) + "\n\n"
		content_write(profile("grabber.log"), content, append=True)

def get_session(netloc):
	s = None
	if netloc not in sessions:
		s = requests.Session()
		s.headers.update(default_header)
		sessions[netloc] = s
	else:
		s = sessions[netloc]
	return s
	

sessions = {}
def grabber(url, header=None, *, referer=None, cookie=None,
		retry=False, done=None, proxy=None, **kwargs):
	"""Request url, return text or bytes of the content."""
	_scheme, netloc, _path, _query, _frag = urlsplit(url)
	
	s = get_session(netloc)
		
	if header:
		s.headers.update(header)

	if referer:
		s.headers['referer'] = quote_unicode(referer)

	if cookie:
		quote_unicode_dict(cookie)
		requests.utils.add_dict_to_cookiejar(s.cookies, cookie)

	if isinstance(proxy, str):
		proxies = {'http': proxy, 'https': proxy}
	else:
		proxies = proxy
		
	r = await_(do_request, s, url, proxies, retry, **kwargs)
	
	if done:
		done(s, r)
	
	return r
	
RETRYABLE_HTTP_CODES = (423, 429, 503)
	
def do_request(s, url, proxies, retry, **kwargs):
	sleep_time = 5
	while True:
		with get_request_lock(url):
			r = s.request(kwargs.pop("method", "GET"), url, timeout=20,
				proxies=proxies, **kwargs)
		grabber_log(url, r.url, r.request.headers, r.headers)

		if r.status_code == 200:
			content_length = r.headers.get("Content-Length")
			if not kwargs.get("stream", False) and content_length and int(content_length) != r.raw.tell():
				raise ValueError(
					"incomplete response. Content-Length: {content_length}, got: {actual}"
						.format(content_length=content_length, actual=r.raw.tell())
				)
			break
		if not retry or r.status_code not in RETRYABLE_HTTP_CODES:
			r.raise_for_status()
		# 302 error without location header
		if r.status_code == 302:
			# pylint: disable=protected-access
			match = re.search(
				r"^location:\s*(.+)",
				str(r.raw._original_response.msg),
				re.M + re.I
			)
			if not match:
				raise TypeError("status 302 without location header")
			url = match.group(1)
			continue
		print(r)
		print("retry after {sleep_time} seconds".format(sleep_time=sleep_time))
		sleep(sleep_time)
		sleep_time *= 2
	return r
	
def grabhtml(*args, **kwargs):
	"""Get html source of given url. Return String."""
	r = grabber(*args, **kwargs)
	guess_encoding(r)
	return r.text

def guess_encoding(r):
	# decode to text
	match = re.search(br"charset=[\"']?([^\"'>]+)", r.content)
	if match:
		encoding = match.group(1).decode("latin-1")
		if encoding == "gb2312":
			encoding = "gbk"
		r.encoding = encoding
	
def _get_ext(r):
	"""Get file extension"""
	b = r.content
	ext = imghdr.what("", b)
	if ext:
		return "." + ext

	# imghdr issue: http://bugs.python.org/issue16512
	if b[:2] == b"\xff\xd8":
		return ".jpg"
		
	# http://www.garykessler.net/library/file_sigs.html
	if b[:4] == b"\x1a\x45\xdf\xa3":
		return ".webm"
		
	if b[:4] == b"RIFF" and b[8:12] == b"WEBP":
		return ".webp"
		
	if b[:4] == b"8BPS":
		return ".psd"
		
	if (b[:16] == b"\x30\x26\xB2\x75\x8E\x66\xCF\x11"
			b"\xA6\xD9\x00\xAA\x00\x62\xCE\x6C"):
		return ".wmv"
		
	mime = None
	if "Content-Type" in r.headers:
		mime = re.search("^(.*?)(;|$)", r.headers["Content-Type"]).group(1)
		mime = mime.strip()
		if "octet-stream" in mime:
			mime = None

	if not mime:
		filename = urlsplit(r.url).path
		mime = puremagic.from_string(b, mime=True, filename=filename)
	
	if mime:
		ext = guess_extension(mime)
		if ext:
			return ext
		# guess_extension doesn't handle video/x-m4v
		match = re.match("\w+/x-(\w+)$", mime)
		if match:
			return f".{match.group(1)}"
		
def get_ext(r):
	"""Get file extension"""
	ext = _get_ext(r)
	# some mapping
	if ext in (".jpeg", ".jpe"):
		return ".jpg"
	return ext

def grabimg(*args, **kwargs):
	"""Grab the image. Return ImgResult"""
	kwargs["stream"] = True
	r = grabber(*args, **kwargs)
	total = int(r.headers.get("Content-Length", "0")) or None
	content_list = []
	with pb_manager.counter(total=total, unit="b", leave=False) as counter:
		for chunk in r.iter_content(2 ** 10):
			content_list.append(chunk)
			counter.update(len(chunk))
	r._content = b"".join(content_list) # pylint: disable=protected-access
	return ImgResult(r)

class ImgResult:
	def __init__(self, response):
		self.response = response
		self.ext = get_ext(response)
		self.bin = response.content



#! python3

from tkinter import ttk
from functools import partial

# https://github.com/PyCQA/pylint/issues/3664
# pylint: disable=dangerous-default-value

class Table:
	def __init__(self, parent, *, tv_opt={}, columns=[]):
		self.sort_mode = None
		self.sort_on = None
		self.listeners = {}
		self.cols = {c["id"]: c for c in columns}
		
		# scrollbar
		scrbar = ttk.Scrollbar(parent)
		scrbar.pack(side="right", fill="y")
		self.scrbar = scrbar
		
		# treeview
		tv = ttk.Treeview(
			parent,
			columns=[c["id"] for c in columns if c["id"] != "#0"],
			yscrollcommand=scrbar.set,
			**tv_opt
		)
		for c in columns:
			tv.heading(c["id"], text=c["text"], command=partial(self.sort_table, id=c["id"]))
			tv.column(c["id"], **{k: v for k, v in c.items() if k in ("width", "anchor")})
		tv.pack(expand=True, fill="both")
		self.tv = tv
		
		scrbar.config(command=tv.yview)
		
		self.key_index = {}
		self.iid_index = {}
		
	def on(self, event_name, listener):
		self.listeners["on_" + event_name] = listener
		
	def sort_table(self, id=None):
		if id == "#0":
			return
		if self.sort_on == id:
			if self.sort_mode == "ASC":
				self.sort_mode = "DESC"
			else:
				self.sort_mode = "ASC"
		else:
			if self.sort_on:
				# reset text
				self.tv.heading(self.sort_on, text=self.cols[self.sort_on]["text"])
			self.sort_mode = self.cols[id].get("sort", "ASC")
			self.sort_on = id
			
		if self.sort_mode == "ASC":
			arrow = "▴"
		else:
			arrow = "▾"
		self.tv.heading(id, text=self.cols[id]["text"] + arrow)
		
		listener = self.listeners.get("on_sort")
		if listener:
			listener(self)
		
	def add(self, row, *, key=None):
		if key and key in self.key_index:
			return
		iid = self.tv.insert("", "end")
		if not key:
			key = iid
		self.key_index[key] = iid
		self.iid_index[iid] = key
		self.update(key, **row)
		return key
		
	def remove(self, *rows):
		self.tv.delete(*[self.key_index[k] for k in rows])
		for key in rows:
			if key not in self.key_index:
				continue
			iid = self.key_index[key]
			del self.key_index[key]
			del self.iid_index[iid]
		
	def clear(self, *, exclude=[]):
		keys = [key for key in self.key_index if key not in exclude]
		self.remove(*keys)
		
	def rearrange(self, rows):
		count = len(self.key_index)
		for key in rows:
			if key not in self.key_index:
				continue
			iid = self.key_index[key]
			self.tv.move(iid, "", count)
		
	def update(self, key, **kwargs):
		if key not in self.key_index:
			return
		iid = self.key_index[key]
		for column, value in kwargs.items():
			self.tv.set(iid, column, value)
			
	def selected(self):
		return [self.iid_index[i] for i in self.tv.selection()]
		
	def contains(self, key):
		return key in self.key_index
		
	def identify_row(self, y):
		iid = self.tv.identify_row(y)
		return self.iid_index.get(iid, None)



#! python3

"""this is danbooru module for comiccrawler

Ex:
	https://danbooru.donmai.us/posts?tags=fault!!

"""

import re
from html import unescape
from urllib.parse import urljoin

from ..core import Episode

domain = ["danbooru.donmai.us"]
name = "Danbooru"
noepfolder = True
config = {
	"curl": "",
	"curl_cdn": ""
}
autocurl = True

def get_title(html, url):
	title = re.search(r"<title>(.+?)</title>", html, re.DOTALL).group(1)
	return title.strip()

def get_episodes(html, url):
	s = []
	for match in re.finditer(r'href="(/posts/(\d+)[^"]*)"', html):
		u = match.group(1)
		title = match.group(2)
		e = Episode(title, urljoin(url, u))
		s.append(e)

	return s[::-1]

def get_images(html, url):
	pos = re.search(r"image-container", html).start()
	img = re.compile(r'data-file-url="([^"]+)"').search(html, pos).group(1)
	return [urljoin(url, img)]

def get_next_page(html, url):
	if re.search("/posts/\d+", url):
		return
		
	m = re.search(r'<a [^>]*rel="next"[^>]*?href="([^"]+)', html)
	if m:		
		return urljoin(url, unescape(m.group(1)))



#! python3

"""
http://ac.qq.com/Comic/comicInfo/id/626619
"""

import re
from deno_vm import eval

from ..core import Episode
from ..grabber import grabhtml
from ..url import urljoin

domain = ["ac.qq.com"]
name = "騰訊"

def get_title(html, url):
	return re.search(r'<h2 class="works-intro-title[^>]+><strong>([^<]+)', html).group(1)

def get_episodes(html, url):
	rx = "class='works-chapter-item'.+?<a[^>]+?title=\"([^\"]+)[^>]+?href=\"([^\"]+)"
	end_index = html.index('class="chapter-page-new')
	episodes = []
	for match in re.finditer(rx, html, re.DOTALL):
		if match.start() >= end_index:
			break
		title, ep_url = match.groups()
		ep_id = re.search("\d+$", ep_url).group()
		title = "{ep_id} - {title}".format(ep_id=ep_id, title=title)
		ep_url = urljoin(url, ep_url)
		episodes.append(Episode(title, ep_url))
	return episodes
	
def get_nonce(html):
	matches = list(re.finditer("window\[.+?=(.+)", html))
	if matches:
		return matches[-1].group(1)
	return re.search("window\.nonce = (.+)", html).group(1)
	
def get_images(html, url):
	data = re.search("var DATA\s*=\s*'[^']+'", html).group()
	nonce = get_nonce(html)
	
	view_js = re.search('src="([^"]+?page\.chapter\.view[^"]+?\.js[^"]*)', html).group(1)
	view_js = grabhtml(urljoin(url, view_js))
	view_js = re.search("(eval\(.+?)\}\(\)", view_js, re.DOTALL).group(1)
	
	code = "\n".join([
		data,
		"""
		function createDummy() {
			return new Proxy(() => true, {
				get: () => createDummy()
			});
		}
		const window = document = createDummy();
		""",
		"const nonce = {};".format(nonce),
		"const W = {DATA, nonce};",
		view_js
	])
	
	data = eval(code)
	return [p["url"] for p in data["picture"]]



#! python3

"""this is acgn module for comiccrawler

Ex:
	http://comic.acgn.cc/manhua-murcielago.htm

"""

import re

from html import unescape

from ..core import Episode
from ..error import SkipEpisodeError

domain = ["comic.acgn.cc"]
name = "ACGN.cc"

def get_title(html, url):
	title = re.search(r"h3><a[^>]+>([^<]+)", html).group(1)
	return unescape(title)

def get_episodes(html, url):
	s = []
	cwd = re.search("https?://[^?]+/", url).group()

	for match in re.finditer(r'a href="(view-[^"]+)"[^>]*>([^<]+)<', html):
		u, title = match.groups()
		s.append(Episode(unescape(title), cwd + u))

	return s

def get_images(html, url):
	return [match.group(1) for match in re.finditer(r'_src="([^"]+)"', html)]

def errorhandler(err, crawler):
	if "暫缺" in crawler.ep.title:
		raise SkipEpisodeError



#! python3

"""山立？

https://www.setnmh.com/comic-lllop-%E5%B1%B1%E6%B5%B7%E9%80%86%E6%88%B0
"""

import re
import json
from html import unescape
from urllib.parse import urljoin

from deno_vm import eval, VM

from ..core import Episode
from ..grabber import grabhtml

domain = ["www.setnmh.com"]
name = "山立"

def get_title(html, url):
	return unescape(re.search('<h1 class="bookname">([^<]+)', html).group(1))

def get_episodes(html, url):
	cartoon_id = re.search('cartoon_id = "([^"]+)', html).group(1)
	code = grabhtml(
		urljoin(url, "/comicinfo-ajaxgetchapter.html"),
		method = "POST",
		data = {
			"cartoon_id": cartoon_id,
			"order_by": "1",
			"chapter_type": "1"
		},
		header = {
			"X-Requested-With": "XMLHttpRequest"
		}
	)
	# breakpoint()
	ep_data = json.loads(code)
	
	ep_url_code = "var moduleExports = value => " + re.search('href = ("/series-.+)', html).group(1)
	with VM(ep_url_code) as make_url:
		s = []
		for item in ep_data["msg"]:
			sys = item["system"]
			s.append(Episode(
				sys["title"],
				urljoin(url, make_url.call("moduleExports", sys))
			))
	return s[::-1]
	
next_page_cache = {}
	
def get_images(html, url):
	if html[0] == '"':
		# wtf
		html = json.loads(html)
	# breakpoint()
	key = re.search(r'var KEY = "([^"]+)', html).group(1)
	cartoon_id = re.search(r'var CARTOON_ID = "([^"]+)', html).group(1)
	chapter_id = re.search(r'var CHAPTER_ID = "([^"]+)', html).group(1)
	page = re.search(r'var PAGE = "([^"]+)', html).group(1)
	total_page = re.search(r'var TOTAL_PAGE = "([^"]+)', html).group(1)
	
	if int(page) < int(total_page):
		next_page_cache[url] = urljoin(url, re.search('href="([^"]+)">下一頁', html).group(1))

	code = grabhtml(
		urljoin(url, "/comicseries/getpictrue.html"),
		method = "POST",
		data = {
			"key": key,
			"cartoon_id": cartoon_id,
			"chapter_id": chapter_id,
			"page": page
		},
		header = {
			"X-Requested-With": "XMLHttpRequest"
		}
	)
	# breakpoint()
	data = eval(code)
	return data["current"]

def get_next_page(html, url):
	if url in next_page_cache:
		return next_page_cache.pop(url)
	



#! python3

# pylint: disable=unused-import
from urllib.parse import urlparse, urlunparse, parse_qs, urlencode, urljoin

URL_PARTS = ("scheme", "netloc", "path", "params", "query", "fragment")

def urlextract(url):
	result = urlparse(url)
	dict = {}
	for prop in URL_PARTS:
		dict[prop] = getattr(result, prop)
	return dict
	
def urlbuild(d):
	return urlunparse(d.get(p) for p in URL_PARTS)

def urlupdate(url, **parts):
	d = urlextract(url)
	d.update(parts)
	return urlbuild(d)
	
def update_qs(url, new_query):
	d = urlextract(url)
	query_dict = parse_qs(d["query"])
	query_dict.update(new_query)
	d["query"] = urlencode(query_dict, doseq=True)
	return urlbuild(d)



#! python3

"""comiccrawler.mods

Import all downloader modules
"""

from os.path import dirname, join, isdir, splitext
from os import listdir
from importlib import import_module
from re import search
from sys import version_info

from ..config import config
from ..profile import get as profile
from ..grabber import cooldown, get_session
from ..util import extract_curl
from ..url import urlparse

def setup_curl(d):
	for key, value in d.items():
		if key.startswith("curl") and value:
			url, headers, cookies = extract_curl(value)
			netloc = urlparse(url).netloc
			s = get_session(netloc)
			for key in list(headers.keys()):
				if key.startswith("If-"):
					headers.pop(key)
			s.headers.update(headers)
			s.cookies.update(cookies)

def import_module_file(ns, file):
	# pylint: disable=import-outside-toplevel
	if version_info < (3, 5):
		from importlib.machinery import SourceFileLoader
		# pylint: disable=deprecated-method, no-value-for-parameter
		module = SourceFileLoader(ns, file).load_module()
	else:
		from importlib.util import spec_from_file_location, module_from_spec
		spec = spec_from_file_location(ns, file)
		module = module_from_spec(spec)
		spec.loader.exec_module(module)
	return module
	
class ModLoader:
	def __init__(self):
		self.mods = set()
		self.domain_index = {}
		self.loaded = False
		
	def load(self):
		"""Load mods"""	
		if self.loaded:
			return
			
		# load from here
		for file in listdir(dirname(__file__)):
			name, ext = splitext(file)
			if name == "__init__":
				continue
			if ext != ".py":
				continue
			self.mods.add(import_module("comiccrawler.mods." + name))
			
		# load from user mods dir
		user_mods_dir = profile("mods")
		if isdir(user_mods_dir):
			for file in listdir(user_mods_dir):
				name, ext = splitext(file)
				if ext != ".py":
					continue
				self.mods.add(import_module_file(
					"comiccrawler.user_mods." + name, join(user_mods_dir, file)))
		
		# build index, update grabber cooldown
		for mod in self.mods:
			for url in mod.domain:
				self.domain_index[url] = mod
			cooldown.update(getattr(mod, "grabber_cooldown", {}))

		# init config
		for mod in self.mods:
			if hasattr(mod, "config"):
				if mod.name not in config.config:
					config.config[mod.name] = {}
				for key, value in mod.config.items():
					if key not in config.config[mod.name]:
						config.config[mod.name][key] = value
			if mod.name not in config.config:
				mod.config = config.config["DEFAULT"]
			else:
				mod.config = config.config[mod.name]
				
		self.load_config()
		self.loaded = True
		
	def list_domain(self, include_mod = False):
		"""Return downloader dictionary."""
		self.load()
		if include_mod:
			return sorted(self.domain_index.items(), key=lambda i: i[0])
		return sorted(self.domain_index)
		
	def get_module(self, url):
		"""Return the downloader mod of spect url or return None"""
		self.load()
		
		match = search(r"^https?://([^/]+?)(:\d+)?/", url)
		
		if not match:
			return None
			
		domain = match.group(1)
		
		while domain:
			if domain in self.domain_index:
				return self.domain_index[domain]
			try:
				domain = domain[domain.index(".") + 1:]
			except ValueError:
				break
		return None
		
	def load_config(self):
		"""Reload config for mods"""
		for mod in self.mods:
			if mod.name in config.config and config.config[mod.name] != mod.config:
				mod.config = config.config[mod.name]
				
			if hasattr(mod, "load_config"):
				mod.load_config()

			if getattr(mod, "autocurl", False):
				setup_curl(mod.config)
	
mod_loader = ModLoader()
list_domain = mod_loader.list_domain
get_module = mod_loader.get_module
load_config = mod_loader.load_config



from requests import HTTPError

class ComicCrawlerSignal(BaseException):
	"""Extend BaseException."""
	pass

class LastPageError(ComicCrawlerSignal):
	"""Raise LastPageError to exit crawl pages loop."""
	pass

class SkipEpisodeError(ComicCrawlerSignal):
	"""Raise SkipEpisodeError to exit crawl episodes loop."""
	def __init__(self, always=True):
		super().__init__()
		self.always = always

class PauseDownloadError(ComicCrawlerSignal):
	"""Raise PauseDownloadError to exit crawl mission loop."""
	pass
	
class SkipPageError(ComicCrawlerSignal):
	"""Raise this error to skip a page in the analyzer."""
	pass

# class ExitErrorLoop(ComicCrawlerSignal):
	# """Raise ExitErrorLoop to exit error loop."""
	# pass

class ComicCrawlerError(Exception):
	"""Extend Exception."""
	pass

class ModuleError(ComicCrawlerError):
	"""Can't find module."""
	pass

def is_403(err):
	return is_http(err, code=403)
	
def is_http(err, code=None):
	if not isinstance(err, HTTPError):
		return False
		
	if code is None:
		return True
		
	try:
		if err.response.status_code == code:
			return True
	except AttributeError:
		pass
		
	return False



"""
https://manga.bilibili.com/detail/mc25940
"""

import json
import re
import io
from urllib.parse import urljoin
from zipfile import ZipFile

from deno_vm import VM

from ..core import Episode
from ..grabber import grabhtml, grabber

domain = ["manga.bilibili.com"]
name = "manga.bilibili"

class ComicDetail(dict):
	def load(self, id):
		result = grabhtml(
			"https://manga.bilibili.com/twirp/comic.v2.Comic/ComicDetail?device=pc&platform=web",
			method="POST",
			json={"comic_id": id}
			)
		result = json.loads(result)
		self[id] = result

comic_detail = ComicDetail()

class Decoder:
	def __init__(self):
		self.module = None

	def loaded(self):
		return bool(self.module)

	def load(self, bili_js_url):
		bili_js = grabhtml(bili_js_url)
		js = """
		const window = {};
		const self = window;
		""" + bili_js + """
		let decode;
		let factory;
		let webpack;
		for (const key in window) {
			if (key.startsWith("webpack")) {
				webpack = window[key];
			}
		}
		for (const fn of Object.values(webpack[0][1])) {
		  if (fn.toString().includes('Indexer')) {
			factory = fn;
			break;
		  }
		}
		const _require = () => ({
		  async loadAsync(data) {
			throw {
			  data,
			  message: 'extract data'
			};
		  }
		});
		_require.d = (exports, o) => {
		  const getValue = Object.values(o)[0];
		  decode = getValue();
		};

		factory({}, {}, _require);

		var exportDecode = (seasonId, episodeId, data) => {
		  return decode(seasonId, episodeId, Buffer.from(data))
			.catch(err => {
				if (err.message !== "extract data") throw err;
				return Array.from(err.data);
			});
		};
		"""
		# import pathlib
		# pathlib.Path("bili.js").write_text(js, encoding="utf8")
		self.module = VM(js)

	def decode(self, id, ep_id, data):
		return bytes(self.module.call('exportDecode', id, ep_id, list(data)))

decoder = Decoder()

def get_title(html, url):
	id = int(re.search(r'mc(\d+)', url).group(1))
	if id not in comic_detail:
		comic_detail.load(id)
	return comic_detail[id]["data"]["title"]

def get_episodes(html, url):
	id = int(re.search(r'mc(\d+)', url).group(1))
	if id not in comic_detail:
		comic_detail.load(id)
	detail = comic_detail.pop(id)
	return [Episode(
		"{} - {}".format(ep["short_title"], ep["title"]),
		urljoin(url, "/mc{}/{}?from=manga_detail".format(id, ep["id"]))
		) for ep in reversed(detail["data"]["ep_list"])]

def get_images(html, url):
	id, ep_id = (int(n) for n in re.search(r'mc(\d+)/(\d+)', url).groups())
	if not decoder.loaded():
		bili_js = re.search(r'src="([^"]+/bili\.\w+\.js)', html).group(1)
		decoder.load(urljoin(url, bili_js))
	image_index = grabhtml(
		"https://manga.bilibili.com/twirp/comic.v1.Comic/GetImageIndex?device=pc&platform=web",
		method="POST",
		json={"ep_id": ep_id}
		)
	image_index = json.loads(image_index)
	image_index_file = grabber(urljoin(image_index["data"]["host"], image_index["data"]["path"])).content
	image_index_file = decoder.decode(id, ep_id, image_index_file)
	index = read_zip(image_index_file, "index.dat")
	index = json.loads(index)
	return [ImageGetter(f) for f in index["pics"]]

def read_zip(b, filename):
	with ZipFile(io.BytesIO(b)) as zip:
		return zip.read(filename).decode("utf8")

class ImageGetter:
	def __init__(self, path):
		self.path = path

	def __call__(self):
		result = grabhtml(
			"https://manga.bilibili.com/twirp/comic.v1.Comic/ImageToken?device=pc&platform=web",
			method="POST",
			json={"urls": json.dumps([self.path])}
			)
		result = json.loads(result)
		return "{}?token={}".format(result["data"][0]["url"], result["data"][0]["token"])




#! python3
import re
from html import unescape

from ..episode import Episode
from ..grabber import grabber, grabhtml
from ..url import urljoin
from ..util import clean_tags, extract_curl

domain = ["fantia.jp"]
name = "fantia"
config = {
	"curl": "",
	"api_curl": ""
}
noepfolder = True
autocurl = True

next_page_cache = {}
pin_entry_cache = {}

def get_title(html, url):
	name = re.search('<h1 class="fanclub-name">(.+?)</h1', html).group(1)
	return f"[fantia] {clean_tags(name)}"
	
def curl_to_kwargs(curl):
	kwargs = {}
	_url, header, cookie = extract_curl(curl)
	# NOTE: method-level header/cookies won't be stored into session
	kwargs["headers"] = header
	kwargs["cookies"] = cookie
	return kwargs

def get_episodes(html, url):
	result = []
	for match in re.finditer('<a[^>]+href="(/posts/(\d+))"[^>]+title="([^"]+)', html):
		ep_url, ep_id, ep_title = match.groups()
		title = f"{ep_id} - {unescape(ep_title)}"
		result.append(Episode(url=urljoin(url, ep_url), title=title))
	return result[::-1]

def get_images(html, url):
	post_id = re.search("posts/(\d+)", url).group(1)
	result = grabber(f"https://fantia.jp/api/v1/posts/{post_id}", **curl_to_kwargs(config["api_curl"])).json()
	thumb = result["post"].get("thumb", {}).get("original")
	if thumb:
		yield thumb
	for content in result["post"]["post_contents"]:
		for photo in content.get("post_content_photos", []):
			i_html = grabhtml(urljoin(url, photo["show_original_uri"]), referer=url)
			yield unescape(re.search('<img src="([^"]+)', i_html).group(1))
		if "download_uri" in content:
			yield urljoin(url, content["download_uri"])
				
def get_next_page(html, url):
	match = re.search('<a rel="next"[^>]*href="([^"]+)', html)
	if match:
		return urljoin(url, match.group(1))



#! python3

"""this is gelbooru module for comiccrawler

Example:
	https://gelbooru.com/index.php?page=post&s=list&tags=hiroyama_hiroshi
	https://gelbooru.com/index.php?page=pool&s=show&id=45250
	https://gelbooru.com/index.php?page=post&s=list&tags=nakajima_ryou

"""

from html import unescape
import re

from ..url import urljoin
from ..core import Episode

domain = ["gelbooru.com"]
name = "Gelbooru"
noepfolder = True

cookie = {
	"fringeBenefits": "yup"
}

def is_pool(url):
	return "page=pool" in url

def get_title(html, url):
	if is_pool(url):
		title = unescape(re.search("<h3>Now Viewing: ([^<]+)", html).group(1))
		pool_id = re.search("id=(\d+)", url).group(1)
		return "[Gelbooru] {title} ({pool_id})".format(title=title, pool_id=pool_id)
	title = unescape(re.search("<title>([^<]+?)\|\s+(?:Gelbooru|Page)", html).group(1).strip())
	return "[Gelbooru] {title}".format(title=title)

def get_episodes(html, url):
	s = []
	for match in re.finditer(r'<a[^>]+?href="((?:(?:https:)?//gelbooru\.com/)?index\.php\?page=post&(?:amp;)?s=view[^"]+)', html):
		ep_url = unescape(match.group(1))
		id = re.search(r"\bid=(\d+)", ep_url).group(1)
		s.append(Episode(id, urljoin(url, ep_url)))
	if is_pool(url):
		return s
	return s[::-1]
	
def get_next_page(html, url):
	paginator_index = html.find('<div id="paginator">')
	if paginator_index >= 0:
		rx = re.compile(r'</b>\s*<a href="([^"]+)')
		match = rx.search(html, paginator_index)
		if match:
			return urljoin(url, unescape(match.group(1)))

def get_images(html, url):
	img = re.search('<a href="([^"]+)"[^>]*>Original image', html).group(1)
	return img
	


#! python3

"""this is 8comic module for comiccrawler.
	
http://www.comicbus.com/html/197.html

"""

import re
from urllib.parse import urljoin

from deno_vm import VM

from ..core import Episode, grabhtml
from ..util import clean_tags
from ..url import update_qs

domain = ["www.8comic.com", "www.comicvip.com", "comicbus.com", "www.comicabc.com"]
name = "無限"
next_page_cache = {}
nview = None

def get_title(html, url):
	return re.search('addhistory\("\d+","([^"]+)',html).group(1)

def get_episodes(html, url):
	html = html.replace("\n", "")
	
	js = """
		var output;
		function getCookie() {}
		function getcookie() {}
		var window = {
			open: function(result){
				output = result;
			}
		};
		var document = {
			location: {
				href: ""
			}
		};
	""" + grabhtml(urljoin(url, "/js/comicview.js"))
	
	s = []
	matches = re.finditer(
		r'<a [^>]*?onclick="(cview[^"]+?);[^>]*>(.+?)</a>',
		html, re.M
	)
	with VM(js) as vm:
		for match in matches:
			cview, title = match.groups()
			
			vm.run(cview)
			ep_url = vm.run("output")
			title = clean_tags(title)

			e = Episode(title, urljoin(url, ep_url))
			s.append(e)
	return s
	
def get_images(html, url):
	global nview
	if not nview:
		nview = re.search('src="([^"]*nview\.js[^"]*)"', html).group(1)
		nview = urljoin(url, nview)
		nview = grabhtml(nview)
	
	try:
		# http://www.comicbus.com/html/103.html
		script = re.search('(var ch=.+?)spp\(\)', html, re.DOTALL).group(1)
	except AttributeError:
		# http://www.comicbus.com/html/7294.html
		script = re.search('(var chs=.+?)</script>', html, re.DOTALL).group(1)
	
	js = """
	var url,
		images = [],
		document = {
			location: {
				toString() {return url;},
				get href() {return url;},
				set href(_url) {url = _url;}
			},
			getElementById() {
				return {
					set src(value) {
						images.push(value);
					},
					style: {}
				};
			}
		},
		navigator = {
			userAgent: "",
			language: ""
		},
		window = {},
		alert = () => {};
		
	function scriptBody() {
		initpage = () => {};
	""" + nview + script + """
		return [images[0], p, ps, ch];
	}
	
	function getImages(url) {
		images = [];
		document.location.href = url;
		return scriptBody();
	}
	"""
	
	with VM(js) as vm:
		img, p, ps, ch = vm.call("getImages", url)
	if p < ps:
		if "/ReadComic/" in url:
			# https://www.comicabc.com/ReadComic/6997/734/734_8d00xI27S.html?p=2
			next_page_cache[url] = update_qs(url, {"p": p + 1})
		else:
			# https://www.comicabc.com/online/new-18117.html?ch=122-2
			next_page_cache[url] = update_qs(url, {"ch": f"{ch}-{p + 1}"})
	

	return urljoin(url, img)

def get_next_page(html, url):
	return next_page_cache.pop(url, None)



#! python3

"""吹雪動漫"""

import re
from urllib.parse import urljoin

from deno_vm import eval

from ..core import Episode
from ..grabber import grabhtml

domain = ["www.chuixue.com", "www.chuixue.net"]
name = "吹雪"

def get_title(html, url):
	return re.search("<h1>([^<]+)", html).group(1)

def get_episodes(html, url):
	id = re.search("manhua/(\d+)", url).group(1)
	s = []
	for match in re.finditer('<a href="(/manhua/' + id + '/\d+\.html)"[^>]*>([^<]+)', html):
		ep_url, title = match.groups()
		s.append(Episode(title, urljoin(url, ep_url)))
	return s[::-1]

global_js = None

def get_images(html, url):
	global global_js
	js = re.search("(var ret_classurl.+?)</script>", html, re.DOTALL).group(1)
	if not global_js:
		global_js_url = re.search('src="([^"]+global\.js)"', html).group(1)
		global_js = grabhtml(urljoin(url, global_js_url))
		global_js = re.search("(var WebimgServer.+?)window\.onerror", global_js, re.DOTALL).group(1)
		
	imgs, server = eval("""
	function request() {
		return "";
	}
	""" + js + global_js + """;
	[photosr.slice(1), WebimgServerURL[0]]
	""")
	return [urljoin(server, img) for img in imgs]



#! python3

"""this is konachan module for tumblr

Example:
	http://otn-pct.tumblr.com/
	https://shonep.tumblr.com/

"""

import re, json
from urllib.parse import urljoin

from ..core import Episode, grabhtml
from ..error import SkipEpisodeError, PauseDownloadError

domain = ["tumblr.com"]
name = "tumblr"
noepfolder = True

config = {
	"full_size": "False",
	"insecure_http": "False",
	"cookie_pfx": "",
	"user_agent": ""
}

header = {}

def load_config():
	header["User-Agent"] = config["user_agent"]

def check_login(html):
	if 'href="https://www.tumblr.com/login"' in html:
		raise PauseDownloadError("You didn't login")

def get_title(html, url):
	title = re.search(r"<title>([^<]+)", html).group(1).strip()
	id = url[url.index("://") + 3: url.index(".tumblr.com")]
	return "[tumblr] {} ({})".format(title, id)
	
def get_episodes(html, url):
	# FIXME: we need a way to check final URL
	check_login(html)
	s = []
	pattern = 'href="(' + re.escape(urljoin(url, "/post/")) + '(\d+)[^"]*)'
	for m in re.finditer(pattern, html):
		url, uid = m.groups()
		s.append(Episode(uid, url))
	return s[::-1]
			
def get_images(html, url):
	images = []
	images.extend(try_get_videos(html, url))
	images.extend(try_get_images(html))
	if not images:
		raise SkipEpisodeError
	return images

def try_get_images(html):
	s = re.search('<script type="application/ld\+json">([^<]*)</script>', html).group(1)
	o = json.loads(s)
	if "image" not in o:
		return
	if isinstance(o["image"], str):
		yield transform(o["image"])
	else:
		yield from (transform(u) for u in o["image"]["@list"])
	
def try_get_videos(html, url):
	match = re.search("<iframe src='(https?://www\.tumblr\.com/video/[^']+)", html)
	if not match:
		return
	frame_src = match.group(1)
	frame_html = grabhtml(frame_src, referer=url)
	source_src = re.search('<source src="([^"]+)', frame_html).group(1)
	source_src = re.sub("/\d+$", "", source_src)
	yield source_src
	
def transform(url):
	"""Map thumbnail to full size"""
	# https://github.com/eight04/ComicCrawler/issues/82
	if config.getboolean("full_size"):
		url = re.sub("[^/]+\.tumblr\.com", "data.tumblr.com", url)
		url = re.sub("_\d+(\.\w+)$", r"_raw\1", url)
	if config.getboolean("insecure_http"):
		url = re.sub("^https:", "http:", url)
	return url

def get_next_page(html, url):
	match = re.search(r'/page/(\d+)', url)
	if match:
		page = int(match.group(1))
	else:
		page = 1
	
	next_url = "/page/{page}".format(page=page + 1)
	
	if next_url in html:
		return urljoin(url, next_url)



#! python3
"""The module for m.dmzj.com

Ex. http://m.dmzj.com/info/qiyuanzui.html
"""

import re
from urllib.parse import urljoin

from deno_vm import eval

from ..core import Episode

domain = ["m.dmzj.com"]
name = "動漫之家M"

def get_title(html, url):
	return re.search('comicName">([^<]+)', html).group(1)

def get_episodes(html, url):
	data_js = re.search("initIntroData(.+?);", html, re.DOTALL).group(1)
	data = eval(data_js)

	ep_data = []
	for category in data:
		ep_data += category["data"]
	ep_data = sorted(ep_data, key=lambda data: data["chapter_order"])

	episodes = []

	for data in ep_data:
		ep_url = "/view/{}/{}.html".format(data["comic_id"], data["id"])
		title = data["title"] + data["chapter_name"]
		episodes.append(Episode(title, urljoin(url, ep_url)))

	return episodes

def get_images(html, url):
	pages_js = re.search(r'page_url":(\[[^\]]+\])', html).group(1)
	pages = eval(pages_js)

	# thumbs.db?!
	# http://manhua.dmzj.com/zhuoyandexiana/3488-20.shtml
	return [page for page in pages if page and not page.lower().endswith("thumbs.db")]



#! python3

"""this is flickr module for comiccrawler

Ex:
	https://www.flickr.com/photos/133767722@N08/
	https://www.flickr.com/photos/dio-tw/

"""

import re
import json
from urllib.parse import urljoin
from html import unescape

from deno_vm import eval

from ..core import Episode, grabhtml
from ..error import is_http, SkipEpisodeError

domain = ["www.flickr.com"]
name = "flickr"
noepfolder = True

def get_title(html, url):
	title = unescape(re.search("<title>([^<]+)</title>", html).group(1)[:-9])
	user = url.partition("/photos/")[2].strip("/")
	return "[flickr] {title} ({user})".format(title=title, user=user)

def get_episodes(html, url):
	eps = []
	
	key = re.search('root.YUI_config.flickr.api.site_key = "([^"]+)', html).group(1)
	nsid = re.search('"nsid":"([^"]+)', html).group(1)
	base = re.match('.+?/photos/[^/]+', url).group()
	match = re.search(r"/page(\d+)", url)
	if match:
		page = int(match.group(1))
	else:
		page = 1
	
	for photo in query_photos(url, key, nsid, page):
		title = "{id} - {title}".format_map(photo)
		ep_url = "{base}/{id}/".format(base=base, id=photo["id"])
		
		if photo.get("media") == "video":
			image = None
		else:
			sizes = get_sizes(photo)
			max_size = max(sizes, key=lambda s: s.get("width", 0))
			image = urljoin(url, max_size["url"])
			
		eps.append(Episode(title, ep_url, image=image))

	return eps[::-1]
	
def get_sizes(photo):
	# photo is a dict
	sizes = {}
	for key, value in photo.items():
		match = re.match("([a-z]+)_([a-z0-9]{1,2})$", key)
		if not match:
			continue
		prop, name = match.groups()
		if name not in sizes:
			sizes[name] = {"name": name}
		sizes[name][prop] = value
	return sizes.values()
	
def query_photos(url, key, nsid, page):
	params = {
		"per_page": 100,
		"page": page,
		"extras": "media,url_sq,url_q,url_t,url_s,url_n,url_w,url_m,url_z,url_c,url_l,url_h,url_k,url_3k,url_4k,url_f,url_5k,url_6k,url_o",
		"api_key": key,
		"format": "json",
		"nojsoncallback": 1
	}
	match = re.search("/(?:sets|albums)/([^/]+)", url)
	if match:
		set_id = match.group(1)
		params["method"] = "flickr.photosets.getPhotos"
		params["photoset_id"] = set_id
	else:
		params["method"] = "flickr.people.getPhotos"
		params["user_id"] = nsid
		
	rs = grabhtml("https://api.flickr.com/services/rest", params=params)
	# pylint: disable=no-member
	# https://github.com/PyCQA/pylint/issues/922
	rs = json.loads(rs)
	return (rs.get("photos") or rs.get("photoset"))["photo"]
	
def query_video(id, secret, key):
	rs = grabhtml("https://api.flickr.com/services/rest", params={
		"photo_id": id,
		"secret": secret,
		"method": "flickr.video.getStreamInfo",
		"api_key": key,
		"format": "json",
		"nojsoncallback": "1"
	})
	rs = json.loads(rs)
	return sorted(rs["streams"]["stream"], key=key_func)[-1]["_content"]
		
prior = {
	"orig": float("inf"),
	"appletv": 720,
	"iphone_wifi": 360
}

def key_func(stream):
	if stream["type"] in prior:
		return prior[stream["type"]]
	if isinstance(stream["type"], str):
		return int(re.match("[\d.]+", stream["type"]).group())
	return stream["type"]
	
def get_images(html, url):
	key = re.search('root\.YUI_config\.flickr\.api\.site_key = "([^"]+)', html).group(1)
	model = re.search(r"Y\.ClientApp\.init\(([\s\S]+?)\)\s*\.then", html).group(1)
	js = """
	const auth = null, reqId = null;
	const model = """ + model + """;
	model.modelExport.main["photo-models"][0]
	"""
	data = eval(js)
	if data.get("mediaType") == "video":
		return query_video(data["id"], data["secret"], key)
	max_size = max(data["sizes"].values(), key=lambda s: s.get("width", 0))
	return urljoin(url, max_size["url"])

def get_next_page(html, url):
	match = re.search('rel="next"\s+href="([^"]+)', html)
	if match:
		return urljoin(url, match.group(1))
		
def errorhandler(err, crawler):
	if is_http(err, 410) or is_http(err, 404):
		if (re.match(r"https://(live|farm\d+)\.staticflickr\.com/\d+/\d+_[a-z0-9]+_[a-z0-9]{1,2}\.\w+", err.response.url) and
				crawler.ep.image):
			# a specific size is deleted?
			crawler.ep.image = None
			# clear html to refetch the page
			crawler.html = None
			return
			
		if re.match(r"https://www\.flickr\.com/photos/[^/]+/\d+/", err.response.url):
			raise SkipEpisodeError



#! python3

"""iibq

Ex:
	http://www.aacomic.com/manhua/blgl20978/

"""

from itertools import cycle
import re

from deno_vm import VM

from ..core import Episode, grabhtml
from ..url import urlparse, urlupdate, urljoin
from ..error import is_http

domain = ["www.aacomic.com"]
name = "暗暗"

def get_title(html, url):
	title = re.search(r"<h1><a.+?>([^<]+)", html, re.DOTALL).group(1)
	return title.strip()
	
def get_episodes(html, url):
	s = []
	html = html[html.index('<div class="cVol'):]
	pattern = r"href='(http://www\.aacomic\.com/comics/\d+viewpage\d+/)'>([^<]+)"
	for match in re.finditer(pattern, html):
		u, title = match.groups()
		s.append(Episode(title, u))
			
	return s[::-1]
	
servers = None

def get_images(html, url):
	s_files = re.search('sFiles="([^"]+)', html).group(1)
	s_path = re.search('sPath="([^"]+)', html).group(1)
	
	viewhtm = re.search(r'src="([^"]*?viewhtm\d*\.js[^"]*)', html)
	viewhtm = grabhtml(urljoin(url, viewhtm.group(1)))
	
	env = """
	window = {
		"eval": eval,
		"parseInt": parseInt,
		"String": String,
		"RegExp": RegExp
	};
	location = {
		"hostname": null
	};
	function setHostname(hostname) {
		location.hostname = hostname;
	}
	"""
	
	js = env + re.search(r'function isMobile\(\){.+?}(.+?)var cMod', viewhtm,
		re.DOTALL).group(1)
	with VM(js) as vm:
		vm.call("setHostname", urlparse(url).hostname)
		arr_files = vm.call("unsuan", s_files).split("|")
	
	ds = re.search(r"src='([^']*?ds\.js[^']*)", html)
	ds = grabhtml(urljoin(url, ds.group(1)))
	
	global servers
	servers = re.search('sDS = "([^"]+)', ds).group(1).split("^")
	servers = [s.split("|")[1] for s in servers]
	servers = cycle(servers)
	server = next(servers)
	return (server + s_path + f for f in arr_files)
	
def errorhandler(err, crawler):
	if not is_http(err):
		return
		
	if not crawler.image or not crawler.image.url:
		return
		
	server = next(servers)
	crawler.image = urlupdate(crawler.image, netloc=server)



"""Comic Crawler

Usage:
  comiccrawler [--profile=<profile>] (
    domains |
    download <url> [--dest=<save_path>] |
    gui
  )
  comiccrawler (--help | --version)

Commands:
  domains             List supported domains.
  download <url>      Download from the URL.
  gui                 Launch TKinter GUI.

Options:
  --profile=<profile> Set profile location. [default: ~/comiccrawler]
  --dest=<save_path>  Set download save path. [default: .]
  --help              Show help message.
  --version           Show current version.

Sub modules:
  comiccrawler.core   Core functions of downloading, analyzing.
  comiccrawler.error  Errors.
  comiccrawler.mods   Import download modules.
"""

from docopt import docopt

from . import __version__

def console_download(url, savepath):
	"""Download url to savepath."""
	# pylint: disable=import-outside-toplevel
	from .mission import Mission
	from .analyzer import Analyzer
	from .crawler import download

	mission = Mission(url=url)
	Analyzer(mission).analyze()
	download(mission, savepath)

def console_init():
	"""Console init."""
	# pylint: disable=import-outside-toplevel
	arguments = docopt(__doc__, version=__version__)
	
	if arguments["--profile"]:
		from .profile import set as set_profile
		set_profile(arguments["--profile"])

	if arguments["domains"]:
		from .mods import list_domain
		print("Supported domains:\n" + ", ".join(list_domain()))

	elif arguments["gui"]:
		from .gui.main_window import MainWindow
		MainWindow()

	elif arguments["download"]:
		console_download(arguments["<url>"], arguments["--dest"])



class Episode:
	"""Create Episode object. Contains information of an episode."""
	def __init__(self, title=None, url=None, current_url=None, current_page=0,
			skip=False, complete=False, total=0, image=None):
		"""Construct."""
		self.title = title
		self.url = url
		self.current_url = current_url
		# position of images on current page, starting from 1.
		# FIXME: should we rename this to current_image_index?
		self.current_page = current_page
		self.skip = skip
		self.complete = complete
		# total number of images in this episode
		self.total = total
		self.image = image



#! python3

"""this is hhxiee module for comiccrawler

Ex:
	http://www.hhxiee.com/comic/1827966/
"""

import re
from urllib.parse import urljoin

from deno_vm import VM

from ..core import Episode, grabhtml

domain = [
	"www.hhxiee.com", "www.hhcomic.cc", "www.hhssee.com", "www.hhmmoo.com",
	"www.hheess.com", "www.1manhua.net"]
name = "汗汗"

def get_title(html, url):
	return re.search("<title>([^<]+)", html).group(1).strip().replace("漫画 - 汗汗漫画", "")
	
def get_episodes(html, url):
	eps = []
	
	for match in re.finditer(r"href='(/page[^>\s']+).+?>([^<]+)", html):
		ep_url, title = match.groups()
		eps.append(Episode(title, urljoin(url, ep_url)))
	return eps[::-1]
	
ctx = None
def build_ctx(url):
	"""Reuse javascript context"""
	global ctx
	js_url = urljoin(url, "/script/view.js")
	js = grabhtml(js_url)
	js = """
		var imgEl = {
				style: {},
				name: ""
			},
			domainEl = {};
			
		location = {
			href: "",
			hostname: ""
		};
		document = {
			location: location,
			getElementById: function(id){
				if (id == "hdDomain") {
					return domainEl;
				}
				if (/^img/.test(id)) {
					return imgEl;
				}
				return {};
			}
		};
		window = {
			document: document,
			eval: eval,
			parseInt: parseInt,
			String: String,
			RegExp: RegExp
		};
		function getImages(url, name, hdDomain) {
			location.href = url;
			location.hostname = url.match(/:\/\/([^\/]+)/)[1];
			imgEl.name = name;
			domainEl.value = hdDomain;
			
			window_onload();
			return imgEl.src;
		}
	""" + js
	ctx = VM(js).create()
			
def get_images(html, url):
	if not ctx:
		build_ctx(url)
		
	hd_domain = re.search('id="hdDomain" value="([^"]+)', html).group(1)
	img_name = re.search('img[^>]+?name="([^"]+)', html).group(1)
	
	return ctx.call("getImages", url, img_name, hd_domain)
	
def get_next_page(html, url):
	try:
		count = int(re.search('id="hdPageCount" value="([^"]+)', html).group(1))
		current = int(re.search('id="hdPageIndex" value="([^"]+)', html).group(1))
	except AttributeError:
		pass
	else:
		if current < count:
			return re.sub(r"/\d+\.html", "/{}.html".format(current + 1), url)



import contextlib
import hashlib
import threading

from .io import json_load, json_dump, move, remove
from .util import safefilepath, debug_log
from .profile import get as profile_get
from .episode import Episode

def get_mission_id(mission):
	"""Use title and sha1 of URL as mission id"""
	return "{title} [{sha1}]".format(
		title=mission.title,
		sha1=hashlib.sha1(mission.url.encode("utf-8")).hexdigest()[:6]
	)
	
@contextlib.contextmanager
def edit_mission_id(mission):
	"""A contextmanager for changing mission title."""
	old_id = get_mission_id(mission)
	try:
		yield
	finally:
		new_id = get_mission_id(mission)
		if old_id != new_id:
			old_path = make_ep_path(old_id)
			new_path = make_ep_path(new_id)
			move(old_path, new_path)
	
def make_ep_path(id):
	"""Construct ep path with id"""
	return profile_get("pool/" + safefilepath(id + ".json"))
	
def get_ep_path(mission):
	"""Return episode save file path"""
	return make_ep_path(get_mission_id(mission))
	
load_episodes_status = {}
load_episodes_lock = threading.Lock()

@contextlib.contextmanager
def load_episodes(mission):
	mission_id = id(mission)
	with load_episodes_lock:
		if not mission.episodes:
			eps = json_load(get_ep_path(mission))
			if eps:
				mission.episodes = [Episode(**e) for e in eps]
		if mission_id in load_episodes_status:
			load_episodes_status[mission_id] += 1
		else:
			load_episodes_status[mission_id] = 1
	debug_log("LOAD_EPISODES")
	try:
		yield
	finally:
		with load_episodes_lock:
			if load_episodes_status[mission_id] == 1:
				if mission.episodes:
					file = get_ep_path(mission)
					json_dump(mission.episodes, file)
					mission.episodes = None
				del load_episodes_status[mission_id]
			else:
				load_episodes_status[mission_id] -= 1
		debug_log("UNLOAD_EPISODES")

def cleanup_episode(mission):
	"""Remove episode save file. (probably because the mission is removed from
	the mission manager)
	"""
	remove(get_ep_path(mission))



#! python3

import re
import subprocess # nosec
import traceback
import platform

# Translate state code to readible text.
STATE = {
	"INIT": "準備",
	"ANALYZED": "解析完成",
	"DOWNLOADING": "下載中",
	"PAUSE": "停止",
	"FINISHED": "完成",
	"ERROR": "錯誤",
	"INTERRUPT": "已刪除",
	"UPDATE": "有更新",
	"ANALYZING": "分析中",
	"ANALYZE_INIT": "準備分析"
}

def safe_tk(text):
	"""Encode U+FFFF+ characters. Tkinter doesn't allow to display these
	character. See:
	http://stackoverflow.com/questions/23530080/how-to-print-non-bmp-unicode-characters-in-tkinter-e-g
	"""

	return re.sub(r"[^\u0000-\uFFFF]", "_", text)
	
def get_scale(root):
	"""To display in high-dpi we need to grab the scale factor from OS"""
	
	# There is no solution on XP
	if platform.system() == "Windows" and platform.release() == "XP":
		return 1.0
	
	# Windows
	# https://github.com/eight04/ComicCrawler/issues/13#issuecomment-229367171
	try:
		# pylint: disable=import-outside-toplevel
		from ctypes import windll
		user32 = windll.user32
		user32.SetProcessDPIAware()
		w = user32.GetSystemMetrics(0)
		return w / root.winfo_screenwidth()
	except ImportError:
		# non-windows
		pass
	except Exception: # pylint: disable=broad-except
		traceback.print_exc()
	
	# GNome
	args = ["gsettings", "get", "org.gnome.desktop.interface", "scaling-factor"]
	try:
		with subprocess.Popen(args, stdout=subprocess.PIPE, # nosec
				universal_newlines=True) as p:
			return float(p.stdout.read().rpartition(" ")[-1])
	except Exception: # pylint: disable=broad-except
		traceback.print_exc()
		
	return 1.0



#! python3

from contextlib import contextmanager
import os
import webbrowser
import sys
from time import time, strftime, localtime

import tkinter as tk
from tkinter import ttk, font, messagebox

import desktop
import worker

from ..mods import list_domain, get_module, load_config
from ..config import setting, config
from ..safeprint import print, printer
from ..mission import create_mission
from ..error import ModuleError
from ..profile import get as profile
from ..util import safefilepath, MIN

from ..download_manager import download_manager
from ..mission_manager import mission_manager
from ..channel import download_ch, mission_ch, message_ch
from ..episode_loader import load_episodes, edit_mission_id

from .table import Table
from .dialog import Dialog
from .core import get_scale, safe_tk, STATE
from .select_episodes import select_episodes

def reselect_episodes(root, mission):
	with load_episodes(mission):
		if select_episodes(root, mission):
			mission.state = "ANALYZED"

def draw_last_update(t):
	if not t:
		return "無"
	return strftime("%Y/%m/%d", localtime(t))

def create_mission_table(parent):
	return Table(
		parent,
		columns = [{
			"id": "#0",
			"text": "#",
			"width": 25
		}, {
			"id": "name",
			"text": "任務"
		}, {
			"id": "host",
			"text": "主機",
			"width": 50,
			"anchor": "center"
		}, {
			"id": "state",
			"text": "狀態",
			"width": 70,
			"anchor": "center"
		}, {
			"id": "last_update",
			"text": "更新",
			"width": 70,
			"anchor": "center",
			"sort": "DESC"
		}]
	)
	
def select_title(parent, mission):
	"""Create dialog to change mission title."""
	class _Dialog(Dialog):
		def create_body(self):
			entry = ttk.Entry(self.body)
			entry.insert(0, safe_tk(mission.title))
			entry.selection_range(0, "end")
			entry.pack()
			entry.focus_set()
			self.entry = entry

		def apply(self):
			return self.entry.get()
			
	new_title = _Dialog(parent, title="重命名").wait()
	if not new_title:
		return
	with edit_mission_id(mission):
		mission.title = new_title

class TkinterLoop:
	def __init__(self, root, do_stuff, delay=100):
		self.root = root
		self.delay = delay
		self.do_stuff = do_stuff
		
	def start(self):
		self.root.after_id = self.root.after(self.delay, self.after)
		self.root.mainloop()
		
	def after(self):
		self.root.after_id = self.root.after(self.delay, self.after)
		self.do_stuff()
		
	@contextmanager
	def pause(self):
		self.root.after_cancel(self.root.after_id)
		try:
			yield
		finally:
			self.root.after_id = self.root.after(self.delay, self.after)
			
class ViewMixin:
	"""Create main window view"""
	def create_view(self):
		"""Draw the window."""
		
		# root
		self.root = tk.Tk()
		self.root.title("Comic Crawler")
		
		# setup theme on linux
		if sys.platform.startswith("linux"):
			try:
				ttk.Style().theme_use("clam")
			except tk.TclError:
				pass
		
		# adjust scale, dimension
		scale = max(get_scale(self.root), 1.0)

		self.root.geometry("{w}x{h}".format(
			w=int(500 * scale),
			h=int(400 * scale)
		))
		
		if scale != 1:
			old_scale = self.root.tk.call('tk', 'scaling')
			self.root.tk.call("tk", "scaling", old_scale * scale)
			
		# Use pt for builtin fonts
		for name in ("TkDefaultFont", "TkTextFont", "TkHeadingFont",
				"TkMenuFont", "TkFixedFont", "TkTooltipFont", "TkCaptionFont",
				"TkSmallCaptionFont", "TkIconFont"):
			f = font.nametofont(name)
			size = f.config()["size"]
			if size < 0:
				size = int(-size / 96 * 72)
				f.config(size=size)

		# Treeview doesn't scale its rowheight
		ttk.Style().configure("Treeview", rowheight=int(20 * scale))
		
		# url label
		tk.Label(self.root,text="輸入連結︰").pack(anchor="w")

		# url entry
		entry_url = ttk.Entry(self.root)
		entry_url.pack(fill="x")
		self.entry_url = entry_url

		# bunch of buttons
		self.create_btn_bar()

		# notebook
		self.notebook = ttk.Notebook(self.root)
		self.notebook.pack(expand=True, fill="both")

		# download manager
		frame = ttk.Frame(self.notebook)
		self.notebook.add(frame, text="任務列表")
		
		# mission table
		self.view_table = create_mission_table(frame)

		# library
		frame = ttk.Frame(self.notebook)
		self.notebook.add(frame, text="圖書館")

		# library buttons
		btn_bar = ttk.Frame(frame)
		btn_bar.pack()

		self.btn_update = ttk.Button(btn_bar, text="檢查更新")
		self.btn_update.pack(side="left")

		self.btn_download_update = ttk.Button(btn_bar, text="下載更新")
		self.btn_download_update.pack(side="left")

		# library treeview scrollbar container
		frame_lib = ttk.Frame(frame)
		frame_lib.pack(expand=True, fill="both")
		
		# library table
		self.library_table = create_mission_table(frame_lib)

		# domain list
		frame = ttk.Frame(self.notebook)
		self.notebook.add(frame, text="支援的網域")
		
		table = Table(frame, columns = [{
			"id": "host",
			"text": "域名"
		}, {
			"id": "mod",
			"text": "模組",
			"anchor": "center"
		}], tv_opt={"show": "headings"})
		
		for domain, module in list_domain(True):
			table.add({
				"host": domain,
				"mod": module.name
			})
			
		# batch analyzer
		frame = ttk.Frame(self.notebook)
		self.notebook.add(frame, text="批次加入")
		
		btn_bar = ttk.Frame(frame)
		btn_bar.pack()
		
		self.btn_batch_analyze = ttk.Button(btn_bar, text="開始分析")
		self.btn_batch_analyze.pack(side="left")
		
		self.btn_batch_analyze_stop = ttk.Button(btn_bar, text="停止分析")
		self.btn_batch_analyze_stop.pack(side="left")
		
		self.text_batch_analyze = create_scrollable_text(frame)
			
		# status bar
		statusbar = ttk.Label(self.root, text="Comic Crawler", anchor="e")
		statusbar.pack(anchor="e")
		self.statusbar = statusbar
		
	def create_btn_bar(self):
		"""Draw the button bar"""
		buttonbox = ttk.Frame(self.root)
		buttonbox.pack()

		btnaddurl = ttk.Button(buttonbox, text="加入連結")
		btnaddurl.pack(side="left")
		self.btn_addurl = btnaddurl

		btnstart = ttk.Button(buttonbox, text="開始下載")
		btnstart.pack(side="left")
		self.btn_start = btnstart

		btnstop = ttk.Button(buttonbox, text="停止下載")
		btnstop.pack(side="left")
		self.btn_stop = btnstop

		btnclean = ttk.Button(buttonbox, text="移除已完成")
		btnclean.pack(side="left")
		self.btn_clean = btnclean

		btnconfig = ttk.Button(buttonbox, text="重載設定檔")
		btnconfig.pack(side="left")
		self.btn_config = btnconfig
	
class EventMixin:
	"""Bind main window events"""
	def bind_event(self):
		"""Bind events."""
		pre_url = None
		
		def trieveclipboard(_event):
			nonlocal pre_url
			# Do nothing if there is something in the entry
			if self.entry_url.get():
				return

			try:
				url = self.root.clipboard_get(type="STRING")
			except tk.TclError:
				return
				
			if "\n" in url:
				return

			if get_module(url) and url != pre_url:
				pre_url = url
				self.entry_url.insert(0, url)
				self.entry_url.selection_range(0, "end")
				self.entry_url.focus_set()
		self.root.bind("<FocusIn>", trieveclipboard)

		def entrykeypress(_event):
			addurl()
		self.entry_url.bind("<Return>", entrykeypress)

		# interface for download manager
		def addurl():
			url = self.entry_url.get()
			self.entry_url.delete(0, "end")
			self.add_url(url)
		self.btn_addurl["command"] = addurl

		def startdownload():
			download_manager.start_download()
		self.btn_start["command"] = startdownload

		def stopdownload():
			download_manager.stop_download()
			print("停止下載")
		self.btn_stop["command"] = stopdownload

		def cleanfinished():
			# mission_manager.clean_finished()
			missions = mission_manager.get_all("view", lambda m: m.state == "FINISHED")
			if not missions:
				return
			mission_manager.remove("view", *missions)
			print("移除 " + ", ".join(mission.title for mission in missions))
		self.btn_clean["command"] = cleanfinished

		def reloadconfig():
			config.load()
			load_config()
			print("設定檔重載成功！")
		self.btn_config["command"] = reloadconfig

		def create_menu_set(name, table):
			"""Create a set of menu"""
			menu = tk.Menu(table.tv, tearoff=False)

			# bind menu helper
			def bind_menu(label):
				def bind_menu_inner(func):
					menu.add_command(label=label, command=func)
					return func
				return bind_menu_inner

			# add commands...
			@bind_menu("刪除")
			def _():
				if self.messagebox("yesno", "Comic Crawler", "確定刪除？"):
					self.remove(name, *table.selected())
					
			if name == "view":
				@bind_menu("刪除（包括圖書館）")
				def _():
					if self.messagebox("yesno", "Comic Crawler", "確定刪除？"):
						selected = table.selected()
						self.remove("view", *selected)
						self.remove("library", *selected)

			@bind_menu("移至頂部")
			def _():
				mission_manager.lift(name, *table.selected())

			@bind_menu("移至底部")
			def _():
				mission_manager.drop(name, *table.selected())

			@bind_menu("改名")
			def _():
				selected = table.selected()
				if not selected:
					return
				mission = selected[0]
				select_title(self.root, mission)

			@bind_menu("重新選擇集數")
			def _():
				for mission in table.selected():
					reselect_episodes(self.root, mission)

			@bind_menu("開啟資料夾")
			def start_explorer(event=None):
				if event:
					mission = table.identify_row(event.y)
					if not mission:
						# click on header
						return
					missions = [mission]
				else:
					missions = table.selected()
					
				for mission in missions:
					savepath = profile(mission.module.config["savepath"])
					folder = os.path.join(savepath, safefilepath(mission.title))
					folder = os.path.expanduser(folder)
					if not os.path.isdir(folder):
						os.makedirs(folder)
					desktop.open(folder)

			@bind_menu("開啟網頁")
			def _():
				for mission in table.selected():
					webbrowser.open(mission.url)

			if name == "view":
				@bind_menu("加入圖書館")
				def _():
					missions = table.selected()
					titles = [ m.title for m in missions ]
					mission_manager.add("library", *missions)
					print("已加入圖書館︰{}".format(", ".join(titles)))
					
			if name == "library":
				@bind_menu("檢查更新")
				def _():
					missions = table.selected()
					download_manager.start_check_update(missions)
					
			# menu call
			def tvmenucall(event):
				menu.tk_popup(event.x_root, event.y_root)
			table.tv.bind("<Button-3>", tvmenucall)

			# dubleclick to start explorer
			table.tv.bind("<Double-Button-1>", start_explorer)
			
			# sort
			def on_sort(table):
				def key_func(m):
					if table.sort_on == "host":
						return m.module.name
					if table.sort_on == "name":
						return m.title
					if table.sort_on == "state":
						return m.state
					if table.sort_on == "last_update":
						return m.last_update or MIN
				mission_manager.sort(name, key_func, reverse=table.sort_mode == "DESC")
			table.on("sort", on_sort)

		create_menu_set("view", self.view_table)

		# library buttons
		def lib_check_update():
			download_manager.start_check_update()
		self.btn_update["command"] = lib_check_update

		def lib_download_update():
			missions = mission_manager.get_all("library", lambda m: m.state == "UPDATE")
			if not missions:
				self.messagebox("error", "Comic Crawler", "沒有新更新的任務")
				return
			mission_manager.add("view", *missions)
			download_manager.start_download()
			self.notebook.select(0)
		self.btn_download_update["command"] = lib_download_update

		# library menu
		create_menu_set("library", self.library_table)
		
		# batch analyze
		def batch_analyze():
			text = self.text_batch_analyze.get("1.0", "end")
			
			try:
				download_manager.start_batch_analyze(text)
			except Exception as err: # pylint: disable=broad-except
				self.messagebox("error", "Comic Crawler", "Failed to batch: {}".format(err))
				return
					
			self.text_batch_analyze.config(state="disabled")
		self.btn_batch_analyze["command"] = batch_analyze
		
		def batch_analyze_stop():
			download_manager.stop_batch_analyze()
		self.btn_batch_analyze_stop["command"] = batch_analyze_stop

		# close window event
		def beforequit():
			if download_manager.is_downloading():
				if not self.messagebox("okcancel", "Comic Crawler", "任務下載中，確定結束？"):
					return
					
			# going to quit
			printer.remove_listener(self.sp_callback)		
		
			self.root.destroy()
			
			download_manager.stop_download()
			download_manager.stop_analyze()
			download_manager.stop_check_update()
			
			mission_manager.save()
			
			config.save()
			
		self.root.protocol("WM_DELETE_WINDOW", beforequit)

class MainWindow(ViewMixin, EventMixin):
	"""Create main window GUI."""
	def __init__(self):
		"""Construct."""
		self.create_view()
		self.bind_event()
		
		self.thread = worker.current()
		self.loop = TkinterLoop(self.root, worker.update)

		self.pool_index = {
			id(mission_manager.view): self.view_table,
			id(mission_manager.library): self.library_table
		}
		
		self.register_listeners()
		
		printer.add_listener(self.sp_callback)

		if (setting.getboolean("libraryautocheck") and
			time() - setting.getfloat("lastcheckupdate", 0) > setting.getfloat("autocheck_interval") * 60 * 60):
			download_manager.start_check_update()
			
		self.update_table(mission_manager.view)
		self.update_table(mission_manager.library)
		
		self.save()
		self.loop.start()
		
	def messagebox(self, type, *args, **kwargs):
		"""Pause the loop when using messagebox"""
		name = None
		if type in ("okcancel", "yesno", "yesnocancel", "retrycancel"):
			name = "ask" + type
		else:
			name = "show" + type
		func = getattr(messagebox, name)
		with self.loop.pause():
			return func(*args, **kwargs)
			
	def save(self):
		"""Save mission periodly"""
		mission_manager.save()
		self.root.after(int(setting.getfloat("autosave", 5) * 60 * 1000), self.save)

	def update_mission_info(self, table, mission):
		"""Update mission info on treeview."""
		if not table.contains(mission):
			return
		table.update(
			mission,
			name=safe_tk(mission.title),
			state=STATE[mission.state],
			last_update=draw_last_update(mission.last_update)
		)

	def register_listeners(self):
		"""Add listeners."""
		
		mission_ch.sub(self.thread)
		download_ch.sub(self.thread)
		message_ch.sub(self.thread)
		
		@self.thread.listen("LOG_MESSAGE")
		def _(event):
			text = event.data.splitlines()[0]
			self.statusbar["text"] = safe_tk(text)

		@self.thread.listen("MISSION_PROPERTY_CHANGED")
		def _(event):
			self.update_mission_info(self.view_table, event.data)
			self.update_mission_info(self.library_table, event.data)

		@self.thread.listen("MISSION_LIST_REARRANGED")
		def _(event):
			self.update_table(event.data)

		@self.thread.listen("MISSION_POOL_LOAD_FAILED")
		def _(event):
			self.messagebox(
				"error",
				"Comic Crawler",
				"讀取存檔失敗！\n{}".format(event.data)
			)

		@self.thread.listen("DOWNLOAD_INVALID")
		def _(event):
			err, mission = event.data
			self.messagebox("error", mission.module.name, err)

		@self.thread.listen("ANALYZE_INVALID")
		def _(event):
			err, mission = event.data
			self.messagebox("error", mission.module.name, err)
			
		@self.thread.listen("LIBRARY_CHECK_UPDATE_FAILED")
		def _(event):
			if hasattr(event.data, "mission"):
				title = event.data.mission.module.name
			else:
				title = "Comic Crawler"
			self.messagebox("error", title, str(event.data))
			
		@self.thread.listen("BATCH_ANALYZE_UPDATE")
		def _(event):
			self.update_batch_text(event.data)
			
		@self.thread.listen("BATCH_ANALYZE_END")
		def _(event):
			err = event.data
			if err and not isinstance(err, worker.WorkerExit):
				self.messagebox("error", "Comic Crawler", "批次加入失敗！{}".format(err))
			self.text_batch_analyze.config(state="normal")
			print("Batch analyze ended")
			
	def update_batch_text(self, missions):
		text = "\n".join(m.url for m in missions)
		self.text_batch_analyze.config(state="normal")
		self.text_batch_analyze.delete("1.0", "end")
		self.text_batch_analyze.insert("1.0", text)
		self.text_batch_analyze.config(state="disabled")
			
	def remove(self, pool_name, *missions):
		"""Wrap mission_manager.remove."""
		for mission in missions:
			if mission.state in ("DOWNLOADING", "ANALYZING"):
				self.messagebox("error", "Comic Crawler", "刪除任務失敗！任務使用中")
		mission_manager.remove(pool_name, *missions)

	def sp_callback(self, text):
		"""Transport text to LOG_MESSAGE event."""
		message_ch.pub("LOG_MESSAGE", text)

	def update_table(self, pool):
		"""Refresh treeview."""
		table = self.pool_index[id(pool)]
		missions = pool.values()
		
		table.clear(exclude=missions)
		
		for mission in missions:
			if not table.contains(mission):
				table.add({
					"name": safe_tk(mission.title),
					"host": mission.module.name,
					"state": STATE[mission.state],
					"last_update": draw_last_update(mission.last_update)
				}, key=mission)
				
		table.rearrange(missions)
		
	def add_analyze(self, mission, on_success=None):
		def on_finished(err):
			if err and not isinstance(err, worker.WorkerExit):
				self.thread.later(
					self.messagebox,
					"error",
					mission.module.name,
					"解析錯誤！\n{}".format(err)
				)
			if not err and on_success:
				on_success()
		download_manager.start_analyze(mission, on_finished=on_finished)
		
	def add_url(self, url):
		try:
			mission = mission_manager.get_by_url(url)
		except KeyError:
			pass
		else:
			conflict_action = mission.module.config.get("mission_conflict_action")
			if conflict_action == "update":
				if self.messagebox(
					"yesno",
					"Comic Crawler",
					safe_tk(mission.title) + "\n\n任務已存在，要檢查更新嗎？",
					default="yes"
				):
					mission.state = 'ANALYZE_INIT'
					self.add_analyze(mission)
			elif conflict_action == "reselect_episodes":
				reselect_episodes(self.root, mission)
			else:
				self.messagebox("error", "Comic Crawler", "任務已存在")
				
			return
		try:
			mission = create_mission(url=url)
		except ModuleError:
			self.messagebox(
				"error",
				"Comic Crawler",
				"建立任務失敗！不支援的網址！"
			)
			return
			
		def on_success():
			if len(mission.episodes) == 1:
				return
				
			if not mission.module.config.getboolean("selectall"):
				for ep in mission.episodes:
					ep.skip = True
					
			# note that on_finished is called in the analyzer thread, we can't
			# call select_episodes directly.
			defer = worker.Defer()
			self.thread.later(select_episodes, self.root, mission, on_closed=defer.resolve)
			if not defer.get():
				mission_manager.remove("view", mission)
				
		self.add_analyze(mission, on_success=on_success)
		
def create_scrollable_text(parent):
	scrbar = ttk.Scrollbar(parent)
	scrbar.pack(side="right", fill="y")
	
	text = tk.Text(parent, height=3, yscrollcommand=scrbar.set)
	text.pack(expand=True, fill="both")
	
	scrbar.config(command=text.yview)

	return text



#! python3

"""Simple io module depressing exceptions"""

import io
import os
from os import path
import pprint
import glob
import time
import shutil
import json
from contextlib import contextmanager, suppress

CHUNK_LIMIT = 500 * 1000 * 1000

def write_big(file, content):
	for piece in range(0, len(content), CHUNK_LIMIT):
		file.write(content[piece:piece + CHUNK_LIMIT])

def exists(file):
	file = path.expanduser(file)
	
	return path.exists(file)

def is_file(file):
	"""Check if the file is file."""
	file = path.expanduser(file)

	return path.isfile(file)

def content_write(file, content, append=False):
	"""Write content to file. Content may be str or bytes."""
	file = path.expanduser(file)

	prepare_folder(path.dirname(file))

	original = file

	if append:
		mode = "a"
	else:
		mode = "w"
		file = file + time.strftime("@%Y-%m-%d_%H%M%S")

	if isinstance(content, bytes):
		mode += "b"
		with io.open(file, mode) as f:
			write_big(f, content)

	else:
		if not isinstance(content, str):
			content = pprint.pformat(content)

		with io.open(file, mode, encoding="utf-8") as f:
			write_big(f, content)

	if file != original:
		os.replace(file, original)

def content_read(file, raw=False):
	"""Read content from file. Return str."""
	file = path.expanduser(file)

	if not path.isfile(file):
		return ""

	if raw:
		with io.open(file, "rb") as f:
			return f.read()
	else:
		with io.open(file, "r", encoding="utf-8-sig") as f:
			return f.read()

def prepare_folder(folder):
	"""If the folder does not exist, create it."""
	folder = path.expanduser(folder)

	if not path.isdir(folder):
		os.makedirs(folder)

	return folder

def prepare_file(file):
	"""If the file does not exist, create it."""
	file = path.expanduser(file)

	prepare_folder(path.dirname(file))

	if not path.isfile(file):
		with io.open(file, "w", encoding="utf8"):
			pass

	return file
	
def move(src, dest):
	"""Move src files to dest. Should support wildcard."""
	src = path.expanduser(src)
	dest = path.expanduser(dest)

	if "*" in src:
		# Wildcard multiple move
		prepare_folder(dest)

		for file in glob.iglob(src):
			os.rename(file, path.join(dest, path.basename(file)))
	else:
		# just a rename
		if not exists(src):
			return

		prepare_folder(path.dirname(dest))
		os.rename(src, dest)

def backup(arg):
	"""Create backup file.
	
	:param arg: str - glob pattern or filename.
	"""
	arg = path.expanduser(arg)
	if "*" in arg:
		# Wildcard multiple copy
		for file in glob.iglob(arg):
			shutil.copyfile(file, file + time.strftime("@%Y-%m-%d_%H%M%S"))
	else:
		if not is_file(arg):
			return
		shutil.copyfile(arg, arg + time.strftime("@%Y-%m-%d_%H%M%S"))

def path_each(folder, callback, mode="f"):
	"""Iter through all file in the folder.

	`folder` may contain "*", or it will use the children in the folder.
	`callback` will recieve a filename param.
	`mode` is a string containing f or/and d, to filter file, directory.
	"""
	folder = path.expanduser(folder)
	if "*" in folder:
		files = glob.iglob(folder)
	elif path.isdir(folder):
		files = [path.join(folder, file) for file in os.listdir(folder)]
	else:
		return

	for file in files:
		if path.isfile(file) and "f" in mode or path.isdir(file) and "d" in mode:
			callback(file)
			
def expand(file):
	"""Expand user"""
	return path.expanduser(file)
	
def dirname(file):
	"""Get dirname"""
	return path.dirname(file)

@contextmanager	
def open(file, mode="r"):
	file = path.expanduser(file)
	prepare_folder(path.dirname(file))
	
	original = file
	if "w" in mode and is_file(file):
		file = file + time.strftime("@%Y-%m-%d_%H%M%S")
		encoding = "utf-8"
	else:
		encoding = "utf-8-sig"
		
	with io.open(file, mode, encoding=encoding) as fp:
		yield fp
	
	if file != original:
		os.replace(file, original)
		
def remove(file):
	file = path.expanduser(file)
	with suppress(FileNotFoundError):
		os.remove(file)

def json_load(file):
	"""My json.load"""
	with suppress(OSError):
		with open(file) as fp:
			return json.load(fp)
				
def json_dump(data, file):
	"""My json.dump"""
	
	def encoder(object):
		"""Encode any object to json."""
		if hasattr(object, "tojson"):
			return object.tojson()
		return vars(object)
		
	with open(file, "w") as fp:
		json.dump(
			data,
			fp,
			indent=4,
			ensure_ascii=False,
			default=encoder
		)



#! python3

"""iibq

Ex:
	http://www.iibq.com/comic/82012136189/

"""

import re

from deno_vm import VM

from ..core import Episode, grabhtml

domain = ["www.iibq.com"]
name = "精明眼"

def get_title(html, url):
	title = re.search(r"<h1[^>]*>(.+?)</h1>", html, re.DOTALL).group(1)
	return title.strip()
	
def get_episodes(html, url):
	s = []
	html = html[html.index('<div class="cVol">'):]
	# base = re.search("(https?://[^/]+)", url).group(1)
	for match in re.finditer(
			r"href='(http://www\.iibq\.com/comic/\d+/viewcomic\d+/)'>([^<]+)",
			html):
		u, title = match.groups()
		s.append(Episode(title, u))
			
	return s[::-1]

def get_images(html, url):
	s_files = re.search('sFiles="([^"]+)"', html).group(1)
	s_path = re.search('sPath="([^"]+)"', html).group(1)
	
	viewhtm = grabhtml("http://www.iibq.com/script/viewhtm.js")
	
	env = """
	window = {
		"eval": eval,
		"parseInt": parseInt,
		"String": String,
		"RegExp": RegExp
	};
	location = {
		"hostname": "www.iibq.com"
	};
	"""
	
	js = env + re.search(r'(.+?)var cuImg', viewhtm, re.DOTALL).group(1)
	with VM(js) as vm:
		arr_files = vm.call("unsuan", s_files).split("|")
	
	ds = grabhtml("http://www.iibq.com/script/ds.js")
	
	sl_url = re.search('sDS = "([^"]+)"', ds).group(1).split("^")[0].split("|")[1]
	
	return [sl_url + s_path + f for f in arr_files]



"""
https://www.tohomh.com/wuquan/
"""

import json
import re
from urllib.parse import urljoin

from ..core import Episode, grabhtml

domain = ["www.tohomh.com", "www.tohomh123.com"]
name = "土豪"

def get_title(html, url):
	return re.search('<h1>([^<]+)', html).group(1)

def get_episodes(html, url):
	base = re.match("https?://[^/]+/([^/]+)", url).group(1)
	pattern = '<a href="(/{}/[^"]+)[^>]+>([^<]+)'.format(base)
	try:
		start = html.index("detail-list-select-1")
	except ValueError:
		start = 0
	try:
		end = html.index("detail-list-select-2")
	except ValueError:
		end = len(html)
	s = []

	for match in re.finditer(pattern, html):
		if match.start() < start:
			continue
		if match.end() > end:
			break
		ep_url, title = match.groups()
		s.append(Episode(title, urljoin(url, ep_url)))
	return s[::-1]

def get_images(html, url):
	did = re.search("did=(\d+)", html).group(1)
	sid = re.search("sid=(\d+)", html).group(1)
	endpoint = urljoin(url, "/action/play/read")
	pcount = int(re.search("pcount = (\d+)", html).group(1))
	
	s = []
	for i in range(pcount):
		s.append(create_image_getter(endpoint, did, sid, i + 1))
	return s
		
def create_image_getter(url, did, sid, iid):
	def get():
		content = grabhtml(url, params={
			"did": did,
			"sid": sid,
			"iid": str(iid)
		})
		return json.loads(content)["Code"]
	return get



#!python3

"""This module provide a global setting object, which will save on exit.
"""

from configparser import ConfigParser
from os.path import expanduser, dirname, isdir, normpath
from os import makedirs

from .profile import get as profile

class CaseSensitiveConfigParser(ConfigParser):
	optionxform = str

class Config:
	default = {
		"savepath": "download",
		"runafterdownload": "",
		"libraryautocheck": "true",
		"autocheck_interval": "24",
		"autosave": "5",
		"errorlog": "false",
		"lastcheckupdate": "0",
		"selectall": "true",
		"mission_conflict_action": "update"
	}
	def __init__(self, path):
		self.path = expanduser(path)
		self.config = CaseSensitiveConfigParser(interpolation=None)
		self.load()
		
	def load(self):
		# this method doesn't raise error
		self.config.read(self.path, 'utf-8-sig')
		
		if "DEFAULT" not in self.config:
			self.config["DEFAULT"] = {}
		
		# backward compatible
		if "ComicCrawler" in self.config:
			self.config["DEFAULT"].update(self.config["ComicCrawler"])
			del self.config["ComicCrawler"]
		
		self.default.update(self.config['DEFAULT'])
		self.config['DEFAULT'].update(self.default)
		
		self.config['DEFAULT']["savepath"] = normpath(self.config['DEFAULT']["savepath"])
		
	def save(self):
		if not isdir(dirname(self.path)):
			makedirs(dirname(self.path))
		with open(self.path, 'w', encoding='utf-8') as f:
			self.config.write(f)
	
config = Config(profile('setting.ini'))
setting = config.config['DEFAULT']



#! python3

"""
https://www.artstation.com/taekwonkim
"""


import json
import math
import re

from deno_vm import eval

from ..url import urljoin, urlparse, parse_qs, update_qs
from ..core import Episode
from ..error import SkipPageError, is_http

domain = ["www.artstation.com"]
name = "ArtStation"
noepfolder = True

EP_PER_PAGE = 50

def get_title(html, url):
	return re.search(r'<title>([^<]*)', html).group(1)
	
def is_project(url):
	return re.search("/users/[^/]+/projects\.json", url)
	
def is_user_home(url):
	return re.search("www\.artstation\.com/[^/]+$", url)
	
def get_episodes(html, url):
	if is_project(url):
		return [Episode(
			title="{id} - {title} ({hash_id})".format_map(d),
			url=d["permalink"],
		) for d in reversed(json.loads(html)["data"])]
	raise SkipPageError

def get_images(html, url):
	match = re.search("/artwork/([^/]+)", url)
	if match:
		hash = match.group(1)
		pattern = "cache\.put\('/projects/{hash}\.json', ('.+?')\);$".format(hash=hash)
		data_json = eval(re.search(pattern, html, re.M).group(1))
	elif re.search("/projects/(\w+)\.json", url):
		data_json = html
	else:
		raise ValueError("Unknown page {}".format(url))
	data = json.loads(data_json)
	return [a["image_url"] for a in data["assets"]]
	
def get_next_page(html, url):
	if is_project(url):
		page = int(parse_qs(urlparse(url).query)["page"][0])
		total_page = math.ceil(json.loads(html)["total_count"] / EP_PER_PAGE)
		return update_qs(url, {"page": page + 1}) if page < total_page else None
			
	if is_user_home(url):
		user = re.search("www\.artstation\.com/([^/]+)", url).group(1)
		return urljoin(url, "/users/{user}/projects.json?page=1".format(user=user))

def errorhandler(err, crawler):
	if not is_http(err, 403) or not crawler.ep.current_url:
		return
	match = re.search("artwork/([^/]+)", crawler.ep.current_url)
	if not match:
		return
	artwork = match.group(1)
	crawler.ep.current_url = urljoin(crawler.ep.current_url, "/projects/{}.json".format(artwork))



"""
https://www.instagram.com/haneame_cos/?hl=zh-tw
"""

import re
import json
from html import unescape

from ..core import Episode
from ..error import is_http, SkipEpisodeError, SkipPageError
from ..url import update_qs
from ..grabber import grabber
from ..util import extract_curl

domain = ["www.instagram.com"]
name = "Instagram"
noepfolder = True

cache_next_page = {}
config = {
	"curl": "",
	"api_curl": ""
}
autocurl = True

def get_title(html, url):
	title = re.search("<title>([^<]+)", html).group(1)
	return "[instagram] {}".format(unescape(title).strip())
	
def grabhandler(grab_method, url, **kwargs):
	if "api/v1" in url:
		return grab_json(url, **kwargs)

def grab_json(url, **kwargs):
	_url, header, cookie = extract_curl(config["api_curl"])
	# NOTE: method-level header/cookies won't be stored into session
	kwargs["headers"] = header
	kwargs["cookies"] = cookie
	return grabber(url, **kwargs).json()

def get_episodes_from_data(data):
	user = data["user"]
	timeline = user["edge_owner_to_timeline_media"]
	eps = []
	for item in timeline["edges"]:
		eps.append(Episode(
			str(item["node"]["shortcode"]),
			"https://www.instagram.com/p/{}/".format(item["node"]["shortcode"])
		))
	end_cursor = None
	if timeline["page_info"]["has_next_page"]:
		end_cursor = timeline["page_info"]["end_cursor"]
	return reversed(eps), end_cursor
	
def get_episodes(html, url):
	if match := re.match(r"https://www\.instagram\.com/([^/]+)/", url):
		username = match.group(1)
		if username != "api":
			next_url = f"https://www.instagram.com/api/v1/feed/user/{username}/username/?count=12"
			cache_next_page[url] = next_url
			raise SkipPageError

	if "api/v1/feed" in url:
		body = html
		result = []
		for item in body["items"]:
			result.append(Episode(
				str(item["code"]),
				f"https://www.instagram.com/p/{item['code']}/"
			))
		
		if result and (next_max_id := body.get("next_max_id", None)):
			next_url = update_qs(url, {"max_id": next_max_id})
			cache_next_page[url] = next_url

		return result[::-1]
		
	raise ValueError("unknown URL: {}".format(url))
	
def get_init_data(html, page):
	shared_data = re.search("window\._sharedData = ([\s\S]+?);</script", html).group(1)
	shared_data = json.loads(shared_data)
	return shared_data["entry_data"][page][0]["graphql"]
	
def get_extra_data(html):
	text = re.search("window\.__additionalDataLoaded\('[^']+',(.*?)\);</script>", html).group(1)
	return json.loads(text)

def find_media(media):
	if media.get("video_versions", None):
		return max(media["video_versions"], key=lambda i: i["height"])["url"]
	return max(media["image_versions2"]["candidates"], key=lambda i: i["height"])["url"]

def extract_json(html, filter=None):
	for match in re.finditer(r'<script type="application/json"[^>]+>([^<]+)</script>', html):
		text = match.group(1)
		if filter and not filter(text):
			continue
		yield json.loads(text)

def extract_json_value(data, key):
	# find object value in nested dict
	if isinstance(data, dict):
		for k, v in data.items():
			if k == key:
				yield v
			else:
				yield from extract_json_value(v, key)
	# find value in list
	elif isinstance(data, list):
		for item in data:
			yield from extract_json_value(item, key)
	
def get_images(html, url):
	result = []
	key = "xdt_api__v1__media__shortcode__web_info"
	for data in extract_json(html, filter=lambda s: key in s):
		for web_info in extract_json_value(data, key):
			for item in web_info["items"]:
				if item.get("carousel_media", None):
					result += [find_media(m) for m in item["carousel_media"]]
				else:
					result.append(find_media(item))
	# FIXME: need a way to detect 404 page
	if not result:
		print("no image found")
		raise SkipEpisodeError(always=True)
	return result

def get_next_page(html, url):
	return cache_next_page.get(url)

def errorhandler(err, crawler):
	if is_http(err, 404) and re.match(r"https://www\.instagram\.com/p/[^/]+/", err.response.url):
		raise SkipEpisodeError(True)
	



#! python3

"""
http://tw.manhuagui.com/comic/25713/
http://tw.manhuagui.com/comic/25713/351280.html#p=1
"""

import re
from itertools import cycle
from urllib.parse import urljoin, urlencode

from deno_vm import VM, eval
from lzstring import LZString

from ..core import Episode, grabhtml

domain = ["seemh.com", "ikanman.com", "manhuagui.com", "www.mhgui.com"]
name = "看漫畫"
config = {
	"nowebp": "False"
}
grabber_cooldown = {
	"www.manhuagui.com": 3,
	"tw.manhuagui.com": 3
}

def get_title(html, url):
	return re.search(r'<h1>([^<]*)', html).group(1)

def get_list(html, cid):
	ep_re = r'href="(/comic/{}/\d+\.html)" title="([^"]+)"|<h4><span>([^<]+)'.format(cid)
	arr = []
	try:
		comment_pos = html.index('id="Comment"')
	except ValueError:
		comment_pos = len(html)

	prefix = ""
	for match in re.finditer(ep_re, html):
		if match.start() >= comment_pos:
			break
		ep_url, title, new_prefix = match.groups()
		if new_prefix:
			prefix = new_prefix
			continue
		arr.append(("{} {}".format(prefix, title), ep_url))
	return arr


def get_episodes(html, url):
	episodes = None
	cid = re.search(r"comic/(\d+)", url).group(1)
	
	# http://tw.ikanman.com/comic/10924/
	episodes = get_list(html, cid)
	
	# http://tw.ikanman.com/comic/4350/
	if not episodes:
		view_state = re.search(
			r'id="__VIEWSTATE" value="([^"]+)', html).group(1)
		ep_html = LZString.decompressFromBase64(view_state)
		episodes = get_list(ep_html, cid)
		
	episodes = [Episode(v[0].strip(), urljoin(url, v[1])) for v in episodes]
	return episodes[::-1]
	
servers = None

def get_images(html, url):
	# build js context
	js = """
	var window = self;
	var cInfo;
	var SMH = {
		imgData: function(data) {
			cInfo = data;
			return {
				preInit: function(){}
			};
		}
	};
	"""
	
	configjs_url = re.search(
		r'src="([^"]+?/config_\w+?\.js)"',
		html
	).group(1)
	configjs = grabhtml(urljoin(url, configjs_url), referer=url)
	js += re.search(
		r'^(var CryptoJS|window\["\\x65\\x76\\x61\\x6c"\]).+',
		configjs,
		re.MULTILINE
	).group()

	js += re.search(
		r'<script type="text/javascript">((eval|window\["\\x65\\x76\\x61\\x6c"\]).+?)</script',
		html
	).group(1)
	
	with VM(js) as vm:
		files, path, params = vm.run("[cInfo.files, cInfo.path, cInfo.sl]")
	
	# find server
	# "http://c.3qfm.com/scripts/core_5C348B32A78647FF4208EACA42FC5F84.js"
	# getpath()
	corejs_url = re.search(
		r'src="([^"]+?/core_\w+?\.js)"',
		html
	).group(1)
	corejs = grabhtml(urljoin(url, corejs_url), referer=url)
	
	# cache server list
	servs = re.search(r"var servs=(.+?),pfuncs=", corejs).group(1)
	servs = eval(servs)
	servs = [host["h"] for category in servs for host in category["hosts"]]
	
	global servers
	servers = cycle(servs)

	host = next(servers)
	
	utils = re.search(r"SMH\.(utils=[\s\S]+?);SMH\.imgData=", corejs).group(1)
	
	js = """
	var location = {
		protocol: "http:"
	};
	""" + utils + """;
	function getFiles(path, files, host) {
		// lets try if it will be faster in javascript
		return files.map(function(file){
			return utils.getPath(host, path + file);
		});
	}
	"""
	with VM(js) as vm:
		images = vm.call("getFiles", path, files, host)
	
	if config.getboolean("nowebp"):
		images = map(lambda i: i[:-5] if i.endswith(".webp") else i, images)
		
	params = urlencode(params)
	images = ["{file}?{params}".format(file=i, params=params) for i in images]
	
	return images
	
def errorhandler(err, crawler):
	"""Change host"""
	if crawler.image and crawler.image.url:
		host = next(servers)
		crawler.image.url = re.sub(
			r"://.+?\.",
			"://{host}.".format(host=host),
			crawler.image.url
		)



#! python3

"""this is sfacg module for comiccrawler

Ex:
	http://comic.sfacg.com/HTML/PFSLL/

"""

import re

from ..core import Episode, grabhtml

domain = ["comic.sfacg.com"]
name = "SF"

def get_title(html, url):
	html = html.replace("\n","")
	t = re.search("<title>(.+?)</title>", html).group(1)
	return t.split(",")[0]

def get_episodes(html, url):
	base = re.search("(https?://[^/]+)", url).group(1)
	s = []
	for m in re.findall("<li><a href=\"(.+?)\" target=\"_blank\">(.+?)</a></li>", html, re.M):
		url, title = m
		title = re.sub("<.+?>","",title)
		e = Episode(title, base + url)
		s.append(e)
	return s[::-1]

def get_images(html, url):
	js = re.search("src=\"(/Utility/.+?\.js)\"", html).group(1)
	base = re.search("(https?://[^/]+)", url).group(1)

	htmljs = grabhtml(base + js)
	# host = "http://coldpic.sfacg.com"
	pics = re.findall("picAy\[\d+\] = \"(.+?)\"", htmljs)
	return [base + pic for pic in pics]



import re
import string
from functools import total_ordering
from pathlib import Path

import uncurl

from .config import setting
from .io import content_write
from .profile import get as profile

def dump(html):
	Path("dump.html").write_text(html, encoding="utf-8")

def extract_curl(cmd):
	if not cmd:
		return
	try:
		context = uncurl.parse_context(cmd)
	except SystemExit:
		raise ValueError(f"Failed parsing curl: {cmd}") from None
	return context.url, context.headers, context.cookies

def create_safefilepath_table():
	table = {}
	table.update({
		"/": "／",
		"\\": "＼",
		"?": "？",
		"|": "｜",
		"<": "＜",
		">": "＞",
		":": "：",
		"\"": "＂",
		"*": "＊"
	})
	table.update({
		c: None for c in set(chr(i) for i in range(128)).difference(string.printable)
	})
	table.update({
		chr(i): " " for i in range(32) if chr(i) not in table
	})
	return str.maketrans(table)
	
safefilepath_table = create_safefilepath_table()
dot_table = str.maketrans({".": "．"})

def safefilepath(s):
	"""Return a safe directory name."""
	s = s.strip().translate(safefilepath_table)
	if s[-1] == ".":
		s = s.translate(dot_table)
	return s

def debug_log(*args):
	if setting.getboolean("errorlog"):
		content_write(profile("debug.log"), ", ".join(args) + "\n", append=True)

def url_extract_filename(url):
	filename = url.rpartition("/")[2]
	filename = re.sub(r"\.\w{3,4}$", "", filename)
	return filename

def clean_tags(html):
	html = re.sub("<script.+?</script>", "", html)
	html = re.sub("<.+?>", "", html)
	html = re.sub("\s+", " ", html)
	return html.strip()

@total_ordering	
class MinimumAny:
	def __le__(self, other):
		return True
		
	def __eq__(self, other):
		return self is other

MIN = MinimumAny()



#! python3

import json
import re

from ..core import Episode, grabhtml
from ..url import urljoin

domain = ["smp.yoedge.com"]
name = "yoedge"

def get_title(html, url):
	return re.search(r"<title>([^<]+)", html).group(1)

def get_episodes(html, url):
	html = html[:html.index("am-topbar-fixed-bottom")]
	s = []
	
	for m in re.finditer(r'<a [^>]*?href="([^"]*?/smp-app/[^"]*)">([^<]+)', html):
		ep_url, title = m.groups()
		s.append(Episode(title, ep_url))
	return s

def get_images(html, url):
	cfg = grabhtml(urljoin(url, "smp_cfg.json"))
	cfg = json.loads(cfg)
	
	pages = [cfg["pages"]["page"][i] for i in cfg["pages"]["order"]]
	return [urljoin(url, p) for p in pages]



#! python3
"""
http://manhua.dmzj.com/fklmfsnyly/

"""

import re
import json
from urllib.parse import urljoin

from deno_vm import eval

from ..core import Episode, grabhtml

domain = ["manhua.dmzj.com"]
name = "動漫之家"

def get_title(html, url):
	return re.search("<h1>(.+?)</h1>", html).group(1)

def get_episodes(html, url):
	result = None
	for method in (get_episodes_html, get_episodes_ajax):
		result = method(html, url)
		if result:
			break
	return result
	
def get_episodes_html(html, url):
	comicurl = re.search("comic_url = \"(.+?)\"", html).group(1)
	pattern = (
		r'href="(/{}[^"]+)" (?: class="color_red")?>(.+?)</a>\s*</li>'
			.format(comicurl)
	)
	s = []
	for match in re.finditer(pattern, html):
		ep_url, title = match.groups()
		s.append(Episode(title, urljoin(url, ep_url)))
	return s
	
def get_episodes_ajax(html, url):
	# http://manhua.dmzj.com/lzsyuan/
	comic_id = re.search('g_comic_id = "([^"]+)', html).group(1)
	data = grabhtml(f'http://api.dmzj.com/dynamic/comicinfo/{comic_id}.json')
	data = json.loads(data)
	s = []
	for chapter in data["data"]["list"]:
		title = chapter["chapter_name"]
		ep_url = urljoin(url, f"{chapter['id']}.shtml?cid={comic_id}")
		s.append(Episode(title, ep_url))
	return reversed(s)
	
def get_images(html, url):
	try:
		return get_images_eval(html, url)
	except AttributeError:
		pass
	return get_images_ajax(html, url)
	
def get_images_ajax(html, url):
	chapter_id, comic_id = re.search(r"(\d+)\.shtml\?cid=(\d+)", url).groups()
	data = grabhtml(f"https://m.dmzj.com/chapinfo/{comic_id}/{chapter_id}.html")
	data = json.loads(data)
	return data["page_url"]

def get_images_eval(html, url):
	# Set base url
	base = "http://images.dmzj.com/"

	# Get urls
	html = html.replace("\n", "")
	s = re.search(r"page = '';\s*(.+?);\s*var g_comic_name", html).group(1)
	pages = eval(s + "; pages")
	pages = eval(pages)

	# thumbs.db?!
	# http://manhua.dmzj.com/zhuoyandexiana/3488-20.shtml
	return [base + page for page in pages if page and not page.lower().endswith("thumbs.db")]



#! python3

"""
https://163.bilibili.com/source/5302242712730021771
"""

import json
import re
from html import unescape
from deno_vm import eval

from ..core import Episode
from ..grabber import grabhtml

domain = ["163.bilibili.com"]
name = "bilibili"

def get_title(html, url):
	title = re.search(r'<h1 class="f-toe sr-detail__heading">([^<]+)', html).group(1)
	return unescape(title.strip())

def get_episodes(html, url):
	id = re.search("source/(\d+)", url).group(1)
	info_url = "https://163.bilibili.com/book/catalog/{}.json".format(id)
	data = json.loads(grabhtml(info_url))
	s = [Episode(
		section["fullTitle"],
		"https://163.bilibili.com/reader/{}/{}".format(id, section["sectionId"])
	) for section in data["catalog"]["sections"][0]["sections"]]
	return s

def get_images(html, url):
	js = re.search(r"(window\.DATA = [\s\S]+?)</script>", html).group(1)
	imgs = eval("""
	const window = {};
	""" + js + """
	window.PG_CONFIG.images.map(i => i.url.slice(0, i.url.length - window.DATA.seedLength));
	""")
	return imgs



#! python3

"""漫畫堆

https://www.manhuadui.com/manhua/huiyedaxiaojiexiangrangwogaobai/
"""

import re
from html import unescape
from urllib.parse import urljoin

from deno_vm import eval

from ..core import Episode
from ..grabber import grabhtml

domain = ["www.manhuadui.com"]
name = "漫畫堆"

def get_title(html, url):
	return unescape(re.search('og:title" content="([^"]+)', html).group(1))

def get_episodes(html, url):
	id = re.search("manhua/([^/]+)", url).group(1)
	s = []
	for match in re.finditer('<a href="(/manhua/' + id + '/\d+\.html)" title="([^"]+)', html):
		ep_url, title = match.groups()
		s.append(Episode(unescape(title), urljoin(url, ep_url)))
	return s
	
class ScriptCache:
	def __init__(self):
		self.cache = {}
		
	def fetch(self, html, url, scripts):
		for script in scripts:
			if script in self.cache:
				continue
			pattern = 'src="([^"]+{})'.format(script)
			js_url = re.search(pattern, html).group(1)
			self.cache[script] = grabhtml(urljoin(url, js_url))
		
	def __str__(self):
		return "\n".join(self.cache.values())
	
scripts = ScriptCache()

def get_images(html, url):
	scripts.fetch(html, url, [
		"crypto-js\.js",
		"decrypt\d+\.js",
		"config\.js",
		"common\.js"
	])
	pre_js = re.search("(var chapterImages.+?)</script>", html, re.DOTALL).group(1)
	main_js = re.search("decrypt\d+\(.+", html).group()
	
	images = eval("""
	function createLocalStorage() {
		const storage = {};
		return {setItem, getItem, removeItem};
		function setItem(key, value) {
			storage[key] = value;
		}
		function getItem(key) {
			return storage[key];
		}
		function removeItem(key) {
			delete storage[key];
		}
	}
	
	const exports = undefined;
	const toastr = {options: {}};
	const top = {
		location: {pathname: ""},
		localStorage: createLocalStorage()
	};
	const jQuery = Object.assign(() => {}, {
		cookie: () => false,
		event: {trigger() {}}
	});
	const $ = jQuery;
	const window = top;
	const SinTheme = {
		initChapter() {},
		getPage() {}
	};
	""" + pre_js + str(scripts) + main_js + """
	const s = [];
	for (let i = 0; i < chapterImages.length; i++) {
		s.push(SinMH.getChapterImage(i + 1));
	}
	s;
	""")
	return images



from time import time

from worker import Worker, WorkerExit, sleep

from .analyzer import Analyzer
from .episode_loader import load_episodes

class BatchAnalyzer:
	def __init__(
		self,
		gen_missions,
		stop_on_error=True,
		on_item_finished=None,
		on_finished=None
	):
		self.thread = Worker(self.analyze)
		self.gen_missions = gen_missions
		self.on_finished = on_finished
		self.on_item_finished = on_item_finished
		self.stop_on_error = stop_on_error
		self.cooldown = {}
		
	def start(self):
		self.thread.start()
		return self
		
	def stop(self):
		self.thread.stop()
		return self
		
	def get_cooldown(self, mission):
		if not hasattr(mission.module, "rest_analyze"):
			return 0
		pre_ts = self.cooldown.get(mission.module.name)
		if pre_ts is None:
			return 0
		cooldown = mission.module.rest_analyze - (time() - pre_ts)
		return cooldown if cooldown > 0 else 0
		
	def analyze(self):
		err = None
		try:
			self.do_analyze()
		except WorkerExit:
			raise
		except BaseException as _err:
			err = _err
			raise
		finally:
			if self.on_finished:
				self.on_finished(err)
			
	def do_analyze(self):
		for mission in self.gen_missions:
			err = None
			try:
				sleep(self.get_cooldown(mission))
				with load_episodes(mission):
					Analyzer(mission).analyze()
			except WorkerExit:
				raise
			except BaseException as _err: # pylint: disable=broad-except
				err = _err
				if self.stop_on_error and (not callable(self.stop_on_error) or self.stop_on_error(err)):
					err.mission = mission
					raise
			finally:
				if self.on_item_finished:
					self.on_item_finished(err, mission)
				self.cooldown[mission.module.name] = time()		



#! python3

"""this is xznj120 module for comiccrawler

Example:
	http://www.xznj120.com/lianai/591/
	http://www.33am.cn/gushi/2026/
	http://www.36rm.cn/lianai/235/

"""

from html import unescape
import re

from deno_vm import eval

from ..url import urljoin
from ..core import Episode, grabhtml

domain = ["www.xznj120.com", "www.33am.cn", "www.36rm.cn"]
name = "受漫畫"

def get_title(html, url):
	return re.search("<h1>([^<]+)", html).group(1)

def get_episodes(html, url):
	s = []
	for match in re.finditer(r'<li><a href="([^"]+)"[^>]+><p>([^<]+)', html):
		ep_url, title = [unescape(t) for t in match.groups()]
		s.append(Episode(title, urljoin(url, ep_url)))
	return s[::-1]
	
def get_images(html, url):
	script = re.search(r'<script>\s*(var qTcms_Cur[\s\S]+?)</script>', html).group(1)
	show_js_src = re.search(r'src="([^"]+?show\.\d+\.js[^"]*)', html).group(1)
	show_js = grabhtml(urljoin(url, show_js_src))
	real_pic_fn = re.search(r'(function f_qTcms_Pic_curUrl_realpic[\s\S]+?)function', show_js).group(1)
	code = """
	{script}
	{real_pic_fn}
	function base64_decode(data) {{
		return Buffer.from(data, "base64").toString();
	}}
	// m.wuyouhui.net/template/wap1/css/d7s/js/show.20170501.js?20190506201115
	Buffer.from(qTcms_S_m_murl_e, "base64")
		.toString()
		.split("$qingtiandy$")
		.filter(u => !/^(--|\+)/.test(u))
		.map(f_qTcms_Pic_curUrl_realpic);
	""".format(script=script, real_pic_fn=real_pic_fn)
	return [urljoin(url, i) for i in eval(code)]



#! python3
"""
https://www.mangacopy.com/comic/{name}
"""

import re
import json
from html import unescape

from Cryptodome.Cipher import AES
from Cryptodome.Util.Padding import unpad

from ..episode import Episode
from ..grabber import grabhtml
from ..url import urljoin

domain = ["copymanga.com", "www.mangacopy.com"]
name = "拷貝"

def get_title(html, url):
	return unescape(re.search("<h6[^>]*>([^<]+)", html).group(1)).strip()

def get_pass(html):
	try:
		return re.search('class="disposablePass"\s+disposable="([^"]+)', html).group(1)
	except AttributeError:
		pass
	try:
		return re.search('class="disPass"\s+contentKey="([^"]+)', html).group(1)
	except AttributeError:
		pass
	return re.search("(?:dio|jojo)\s*=\s*'([^']+)", html).group(1)

def get_data(html):
	try:
		return re.search('class="disposableData"\s+disposable="([^"]+)', html).group(1)
	except AttributeError:
		pass
	try:
		return re.search('class="disData"\s+contentKey="([^"]+)', html).group(1)	
	except AttributeError:
		pass
	return re.search('class="imageData"\s+contentKey="([^"]+)', html).group(1)
	
def extract_data(data, pa):
	pa = pa.encode("utf8")
	iv = data[:16].encode("utf8")
	
	cipher = AES.new(pa, AES.MODE_CBC, iv=iv)
	result = unpad(cipher.decrypt(bytes.fromhex(data[16:])), AES.block_size).decode("utf8")
	return json.loads(result)
	

def get_episodes(html, url):
	cid = re.search("[^/]+$", url).group()
	data = json.loads(grabhtml(urljoin(url, f"/comicdetail/{cid}/chapters")))["results"]
	data = extract_data(data, get_pass(html))
	for group in data["groups"].values():
		for chapter in group["chapters"]:
			yield Episode(
				title=chapter["name"],
				url=f"{url}/chapter/{chapter['id']}"
			)
	
def get_images(html, url):
	data = extract_data(get_data(html), get_pass(html))
	return [urljoin(url, i["url"]) for i in data]
	



#! python3

"""this is weibo module for comiccrawler

Ex:
	http://tw.weibo.com/2362470852/4061673218817198
	http://tw.weibo.com/wulazula

"""

import re

from ..core import Episode
from ..url import urljoin
from ..error import SkipEpisodeError, is_http

domain = ["tw.weibo.com"]
name = "weibo"
noepfolder = True

def get_title(html, url):
	match = re.search("^http://tw\.weibo\.com/(\w+)$", url)
	if match:
		user_id = match.group(1)
		user_name = re.search('class="name">\s*<h3><a[^>]+>([^<]+)', html).group(1)
		return "[weibo] {user_id} - {user_name}".format(user_id=user_id, user_name=user_name)
	title = re.search('<title>([^<]+)', html).group(1)
	title = title.replace("- 微博精選 - 微博台灣站", "").replace("\n", "").strip()
	id = re.search('tw\.weibo\.com/([^?]+)', url).group(1)
	return "[weibo] {title} ({id})".format(title=title, id=id)

def get_episodes(html, url):
	if re.match("http://tw\.weibo\.com/\w+/\d+$", url):
		return [Episode("image", url)]
		
	s = []
	pattern = 'class="img_link" href="(http://tw.weibo.com/(\w+/\d+))">'
	for match in re.finditer(pattern, html):
		ep_url, ep_title = match.groups()
		s.append(Episode(ep_title, ep_url))
	return s[::-1]

def get_images(html, url):
	images = re.findall('<img src="([^"]+?sinaimg\.cn/bmiddle[^"]+)', html)
	return map(lambda i: i.replace("/bmiddle/", "/large/"), images)
	
def get_next_page(html, url):
	match = re.search('class="pgNext"><a href="([^"]+)"', html)
	if match:
		return urljoin(url, match.group(1))
		
def errorhandler(err, crawler):
	if is_http(err, 404) and re.search("weibo\.com/[^/]+/[^/]+$", err.request.url):
		raise SkipEpisodeError



import os.path

from .io import path_each
from .util import safefilepath

class SavePath:
	def __init__(self, root, mission, ep, escape=safefilepath):
		self.root = root
		self.mission_title = escape(mission.title)
		self.ep_title = escape(ep.title)
		self.noepfolder = mission.module.config.getboolean(
			"noepfolder",
			fallback=getattr(mission.module, "noepfolder", False)
		)
		self.files = None
		self.escape = escape
		
	def parent(self):
		if self.noepfolder:
			return os.path.join(self.root, self.mission_title)
		return os.path.join(self.root, self.mission_title, self.ep_title)
		
	def filename(self, page, ext=""):
		"""Build filename with page and ext"""
		if not isinstance(page, str):
			page = "{:03d}".format(page)
			
		page = self.escape(page)
			
		if self.noepfolder:
			return "{ep_title}_{page}{ext}".format(
				ep_title=self.ep_title,
				page=page,
				ext=ext
			)
		return "{page}{ext}".format(
			page=page,
			ext=ext
		)
		
	def full_fn(self, page, ext=""):
		"""Build full filename with page and ext"""
		return os.path.join(self.parent(), self.filename(page, ext))

	def exists(self, page):
		"""Check if current page exists in savepath."""
		if page is None:
			return False
		
		# FIXME: if multiple SavePath is created and sharing same .parent(), 
		# they should share the .files too.
		if self.files is None:
		
			self.files = {}
			
			def build_file_table(file):
				_dir, name = os.path.split(file)
				base, ext = os.path.splitext(name)
				self.files[base] = ext
				
			path_each(
				self.parent(),
				build_file_table
			)
			
		ext = self.files.get(self.filename(page))
		return ext and "@" not in ext



#! python3

"""this is mh160 module for comiccrawler

https://www.mh160.com/kanmanhua/30526/
"""

import re
from html import unescape
from urllib.parse import urlparse, urljoin
from deno_vm import eval

from ..core import Episode
from ..grabber import grabhtml

domain = ["www.mh160.com"]
name = "160"

def get_title(html, url):
	match = re.search('<meta property="og:title" content="([^"]+)', html)
	return unescape(match.group(1))

def get_episodes(html, url):
	path = urlparse(url).path
	pattern = '<a href="({}[^"]+)" title="([^"]+)'.format(path)
	s = []
	for match in re.finditer(pattern, html):
		ep_url, ep_title = match.groups()
		s.append(Episode(unescape(ep_title), urljoin(url, ep_url)))
	return s[::-1]
	
def get_images(html, url):
	js_url = re.search(r'src="([^"]+base64\.js)"', html).group(1)
	js_content = grabhtml(urljoin(url, js_url))
	data = re.search('(var chapterTree=.+?)</script>', html, re.DOTALL).group(1)
	match = re.search(r'window\["\\x65\\x76\\x61\\x6c"\](.+?)</script>', html, re.DOTALL)
	data2 = match.group(1) if match else ""

	imgs = eval("""
	const document = {{}};
	{};
	{};
	eval({});
	getUrlpics().map(getrealurl);
	""".format(js_content, data, data2))
	
	return imgs



#! python3

from os import getcwd
from os.path import expanduser, normpath, join

PROFILE = "~/comiccrawler"

def set(profile):
	global PROFILE
	PROFILE = profile

def get(file=None):
	if file is not None:
		return normpath(join(getcwd(), expanduser(PROFILE), expanduser(file)))
	return normpath(join(getcwd(), expanduser(PROFILE)))



import re
import time
import traceback

from worker import WorkerExit, sleep

from .channel import download_ch
from .module_grabber import ModuleGrabber
from .error import SkipPageError, PauseDownloadError, LastPageError
from .safeprint import print

class EpisodeList:
	def __init__(self, eps=()):
		self.list = []
		self.title_set = set()
		self.url_set = set()
		for ep in eps:
			self.add(ep)
	
	def add(self, ep):
		if ep in self:
			return False
		self.list.append(ep)
		self.url_set.add(ep.url)
		self.title_set.add(ep.title)
		return True
		
	def __contains__(self, ep):
		if ep.url in self.url_set:
			return True
		if ep.title in self.title_set:
			return True
		return False
		
	def __iter__(self):
		return iter(self.list)
		
	def __len__(self):
		return len(self.list)
		
	def __reversed__(self):
		return reversed(self.list)

class Analyzer:
	"""Analyze mission"""
	def __init__(self, mission):
		self.mission = mission
		self.grabber = ModuleGrabber(mission.module)
		self.old_urls = None
		self.old_titles = None
		self.is_new = not mission.episodes
		self.html = None
		
	def analyze(self):
		"""Start analyze"""
		try:
			self.do_analyze()
		except BaseException:
			# correctly handle mission state
			self.mission.state = "ERROR"
			raise
		
	def do_analyze(self):
		"""Analyze inner"""
		print("Start analyzing {}".format(self.mission.url))

		self.mission.state = "ANALYZING"
		
		# one-time mission
		if self.is_onetime():
			print("It's one-time mission")
			ep = self.mission.episodes[0]
			if ep.skip or ep.complete:
				self.mission.state = "FINISHED"
			else:
				self.mission.state = "UPDATE"
			print("Analyzing success!")
			return
			
		self.html = self.grabber.html(self.mission.url, retry=True)
		
		if not self.mission.title:
			self.mission.title = self.mission.module.get_title(
				self.html, self.mission.url)
				
		self.analyze_pages()
				
		if self.is_new:
			self.mission.state = "ANALYZED"
			
		elif all(e.complete or e.skip for e in self.mission.episodes):
			self.mission.state = "FINISHED"
			
		else:
			self.mission.state = "UPDATE"

		print("Analyzing success!")
		
	def analyze_pages(self):
		"""Crawl for each pages"""
		url = self.mission.url
		old_eps = EpisodeList(self.mission.episodes or ())
		new_eps = EpisodeList()
		
		while True:
			try:
				eps = list(self.mission.module.get_episodes(self.html, url))
			except SkipPageError:
				pass
			except LastPageError:
				break
			else:
				if not eps:
					print("Warning: get_episodes returns an empty list")
				self.transform_title(eps)
				
				eps = EpisodeList(eps)
				
				# add result episodes into new_eps in new to old order.
				for ep in reversed(eps):
					new_eps.add(ep)
					
				# FIXME: do we really need this check?
				# one-time mission?
				if self.is_onetime(new_eps):
					break
					
				# duplicate with old_eps
				if any(e in old_eps for e in eps):
					break
				
			# get next page
			next_url = self.get_next_page(self.html, url)
			if not next_url:
				break
			url = next_url
			print('Analyzing {}...'.format(url))
			sleep(getattr(self.mission.module, "rest_analyze", 0))
			self.html = self.grabber.html(url, retry=True)
		
		has_new_ep = False
		for ep in reversed(new_eps):
			if old_eps.add(ep):
				has_new_ep = True
		self.mission.episodes = list(old_eps)
		
		if has_new_ep:
			self.mission.last_update = time.time()
		
		if not self.mission.episodes:
			raise TypeError("Episode list is empty")
			
	def get_next_page(self, html, url):
		if not hasattr(self.mission.module, "get_next_page"):
			return None
		return self.mission.module.get_next_page(html, url)
			
	def transform_title(self, eps):
		format = self.mission.module.config.get("titlenumberformat")
		if not format:
			return
		for ep in eps:
			# ignore mission title if exists
			title = list(ep.title.partition(self.mission.title))
			for i in (0, 2):
				title[i] = format_title_number(title[i], format)
			ep.title = "".join(title)
			
	def is_onetime(self, it=None):
		"""Check if the mission should only be analyze once"""
		if it is None:
			it = self.mission.episodes
		return it and len(it) and first(it).url == self.mission.url

def analyze(mission):
	"""Analyze mission."""
	try:
		Analyzer(mission).analyze()

	except WorkerExit: # pylint: disable=try-except-raise
		raise

	except Exception as err: # pylint: disable=broad-except
		traceback.print_exc()
		download_ch.pub("ANALYZE_FAILED", (err, mission))

	except PauseDownloadError as err:
		download_ch.pub("ANALYZE_INVALID", (err, mission))

	else:
		download_ch.pub("ANALYZE_FINISHED", mission)
		
def format_title_number(title, format):
	"""第3卷 --> 第003卷"""
	def replacer(match):
		number = match.group()
		return format.format(int(number))
	return re.sub(r"\d+", replacer, title)

def first(s):
	return next(iter(s))
	



#! python3

"""this is nico seiga module for comiccrawler

Ex:
	http://seiga.nicovideo.jp/user/illust/11937543?target=illust_all

"""

import re

from html import unescape

from urllib.parse import urljoin

from ..core import Episode, grabhtml
from ..error import PauseDownloadError

cookie = {
	"skip_fetish_warning": "1"
}
domain = ["seiga.nicovideo.jp"]
name = "Nico"
noepfolder = True
config = {
	"cookie_user_session": "請輸入Cookie中的user_session"
}

def get_title(html, url):
	if "/user/" in url:
		artist = re.search(r'nickname">([^<]+)', html).group(1)
		id = re.search(r'data-id="(\d+)', html).group(1)
		return "[Nico] {} - {}".format(id, artist)
		
	title = re.search("<title>([^<]+)", html).group(1).partition(" - ")[0]
	return "[Nico] {}".format(title)

def get_episodes(html, url):
	s = []
	for match in re.finditer(r'href="(/seiga/im\d+)">\s*<span[^>]*><img[^>]*?alt="([^"]*)', html):
		ep_url, title = match.groups()
		id = ep_url[7:]
		e = Episode("{} - {}".format(id, unescape(title)), urljoin(url, ep_url))
		s.append(e)
	return s[::-1]

def get_images(html, url):
	if "<!-- ▼Login -->" in html:
		raise PauseDownloadError("You didn't login!")

	source_url = re.search(r'href="(/image/source/\d+)', html)
	if source_url:
		# FIXME: The image is downloaded twice within redirect!
		source_url = urljoin(url, source_url.group(1))
		source_html = grabhtml(source_url)
		image = re.search(r'data-src="([^"]+)', source_html)
		if image:
			image = image.group(1)
		else:
			image = source_url
		return [image]

	img = re.search(r'href="(/image/source\?id=\d+)', html).group(1)
	return [urljoin(url, img)]

def get_next_page(html, url):
	match = re.search(r'href="([^"]+)" rel="next"', html)
	if match:
		return urljoin(url, unescape(match.group(1)))



#! python3

"""this is tuchong module for comiccrawler

Ex:
	https://tuchong.com/\d+/

"""

import re

from ..core import Episode

domain = ["tuchong.com"]
name = "圖蟲"

def get_title(html, url):
	author = re.search('meta name="author" content="([^"]+)', html).group(1)
	id = re.search('tuchong\.com/(\d+)', url).group(1)
	return "[圖蟲] {author} ({id})".format(author=author, id=id)

def get_episodes(html, url):
	s = []

	for match in re.finditer('href="([^"]+?tuchong\.com/\d+/(\d+)[^"]*)" title="([^"]+)', html):
		ep_url, ep_id, title = match.groups()
		title = "{id} - {title}".format(id=ep_id, title=title)
		s.append(Episode(title, ep_url))

	return s[::-1]

def get_images(html, url):
	return re.findall('<img src="([^"]+?photo\.tuchong\.com[^"]+)', html)



#! python3

import re

from urllib.parse import urljoin

from ..core import Episode

domain = ["imgbox.com"]
name = "imgbox"
noepfolder = True

def get_title(html, url):
	title = re.search(r'<h1>([^<]+)</h1>', html).group(1)
	title = re.sub(r" - \d+ images$", "", title)
	return "[imgbox] " + title

def get_episodes(html, url):
	s = []
	for match in re.finditer(r'href="(/(\w+))"><img', html):
		ep_url, id = match.groups()
		s.append(Episode(id, urljoin(url, ep_url)))
	return s

def get_images(html, url):
	img = re.search(r'href="([^?"]+\?download=true)"', html).group(1)
	return [img]



#! python3

"""Mission Manager"""

from collections import OrderedDict
from threading import Lock

from worker import current

from .safeprint import print
from .mission import create_mission, mission_lock
from .episode import Episode
from .io import backup, json_load, json_dump
from .profile import get as profile
from .channel import mission_ch
from .episode_loader import cleanup_episode

class MissionManager:
	"""Since check_update thread might grab mission from mission_manager, we
	have to make it thread safe.
	"""
	def __init__(self):
		"""Construct."""
		self.pool = {}
		self.view = OrderedDict()
		self.library = OrderedDict()
		self.edit = False
		self.lock = Lock()
		
		self.load()

		thread = current()
		mission_ch.sub(thread)
		@thread.listen("MISSION_PROPERTY_CHANGED")
		def _(event):
			"""Set the edit flag after mission changed."""
			self.edit = True

	def cleanup(self):
		"""Cleanup unused missions"""
		main_pool = set(self.pool)
		view_pool = set(self.view)
		library_pool = set(self.library)

		for url in main_pool - (view_pool | library_pool):
			cleanup_episode(self.pool[url])
			del self.pool[url]

	def save(self):
		"""Save missions to json."""
		if not self.edit:
			return

		with mission_lock:
			json_dump(list(self.pool.values()), profile("pool.json"))
			json_dump(list(self.view), profile("view.json"))
			json_dump(list(self.library), profile("library.json"))
			
		self.edit = False
		print("Session saved")

	def load(self):
		"""Load mission from json.

		If failing to load missions, create json backup .
		"""
		try:
			self._load()
		except Exception:
			print("Failed to load session!")
			backup(profile("*.json"))
			raise
		self.cleanup()

	def _load(self):
		"""Load missions from json. Called by MissionManager.load."""
		pool = json_load(profile("pool.json")) or []
		view = json_load(profile("view.json")) or []
		library = json_load(profile("library.json")) or []

		for m_data in pool:
			# reset state
			if m_data["state"] in ("DOWNLOADING", "ANALYZING"):
				m_data["state"] = "ERROR"
			# build episodes
			# compatible 2016.6.4
			if m_data["episodes"]:
				episodes = []
				for ep_data in m_data["episodes"]:
					# compatible 2016.4.3
					if "total" not in ep_data:
						if not ep_data["current_url"]:
							ep_data["total"] = 0
							
						elif ep_data["url"] == ep_data["current_url"]:
							# first page crawler
							ep_data["total"] = ep_data["current_page"] - 1
							
						else:
							# per page crawler
							ep_data["total"] = ep_data["current_page"] - 1
							ep_data["current_page"] = 1
						
						if ep_data["complete"]:
							ep_data["total"] += 1
							
					episodes.append(Episode(**ep_data))
				m_data["episodes"] = episodes
			mission = create_mission(**m_data)
			
			self.pool[mission.url] = mission

		for url in view:
			self.view[url] = self.pool[url]

		for url in library:
			self.library[url] = self.pool[url]

		mission_ch.pub("MISSION_LIST_REARRANGED", self.view)
		mission_ch.pub("MISSION_LIST_REARRANGED", self.library)

	def add(self, pool_name, *missions):
		"""Add missions to pool."""
		pool = getattr(self, pool_name)

		with self.lock:
			for mission in missions:
				if mission.url not in self.pool:
					mission_ch.pub("MISSION_ADDED", mission)
				self.pool[mission.url] = mission					
				pool[mission.url] = mission
		
		mission_ch.pub("MISSION_LIST_REARRANGED", pool)
		self.edit = True

	def remove(self, pool_name, *missions):
		"""Remove missions from pool."""
		pool = getattr(self, pool_name)

		# check mission state
		missions = [m for m in missions if m.state not in ("ANALYZING", "DOWNLOADING")]

		with self.lock:
			for mission in missions:
				if mission.url in pool:
					del pool[mission.url]
			self.cleanup()
			
		mission_ch.pub("MISSION_LIST_REARRANGED", pool)
		self.edit = True

	def lift(self, pool_name, *missions):
		"""Lift missions to the top."""
		pool = getattr(self, pool_name)
		with self.lock:
			for mission in reversed(missions):
				pool.move_to_end(mission.url, last=False)
		mission_ch.pub("MISSION_LIST_REARRANGED", pool)
		self.edit = True

	def drop(self, pool_name, *missions):
		"""Drop missions to the bottom."""
		pool = getattr(self, pool_name)
		with self.lock:
			for mission in missions:
				pool.move_to_end(mission.url)
		mission_ch.pub("MISSION_LIST_REARRANGED", pool)
		self.edit = True
		
	def sort(self, pool_name, key, reverse=False):
		pool = getattr(self, pool_name)
		with self.lock:
			for mission in sorted(pool.values(), key=key):
				pool.move_to_end(mission.url, last=not reverse)
		mission_ch.pub("MISSION_LIST_REARRANGED", pool)
		self.edit = True
		
	def get_all(self, pool_name, test=None):
		"""Get all missions matching condition."""
		with self.lock:
			return [m for m in getattr(self, pool_name).values() if not test or test(m)]
			
	def get(self, pool_name, test=None):
		"""Get the first mission matching condition."""
		with self.lock:
			for mission in getattr(self, pool_name).values():
				if not test or test(mission):
					return mission

	def get_by_url(self, url, pool_name=None):
		"""Get mission by url."""
		if not pool_name:
			return self.pool[url]
		return getattr(self, pool_name)[url]

mission_manager = MissionManager()



#! python3

"""
https://linevoom.line.me/user/*
"""

import re, json
from datetime import datetime
from html import unescape
from ..core import Episode
from ..image import Image
from ..url import update_qs
from ..error import LastPageError

domain = ["linevoom.line.me"]
name = "LINE VOOM"
noepfolder = True
config = {
	"curl": ""
	}
autocurl = True
next_page_cache = {}

def get_title(html, url):
	title = re.search("<title[^>]*>([^<]+)", html).group(1)
	title = title.split("|")[0].strip()
	return unescape(f"[LINEVOOM] {title}")

def get_episodes_from_posts(posts):
	result = []
	next_page = "https://linevoom.line.me/api/socialprofile/getPosts?homeId=_ddkO_XiDqCX3mwNJX6Xu6KzM9Qv7Vrvo-uI0UWU&withSocialHomeInfo=false&postId=1168815423900551283&updatedTime=1688154239000&postLimit=10&commentLimit=0&likeLimit=0"
	for post in posts:
		home_id = post["postInfo"]["homeId"]
		id = post["postInfo"]["postId"]
		created_time = post["postInfo"]["createdTime"]
		updated_time = post["postInfo"]["updatedTime"]
		# it is possible that the post has only media
		text = post["contents"].get("text", "")
		images = []
		images.append(Image(
			data={
				"created_time": created_time,
				"text": text
				}
			))
		images.extend(get_medias(post["contents"]["media"]))
		images.extend(get_bound_countent(post["contents"]))
		ep = Episode(
			title=id,
			url=f"https://linevoom.line.me/post/{id}",
			image=images
			)
		next_page = update_qs(next_page, {"homeId": home_id, "postId": id, "updatedTime": updated_time})
		result.append(ep)
	return result, next_page

def get_bound_countent(contents):
	try:
		media = contents["boundContent"]["data"]["media"]["watermark"]["hashedProxyValue"]
	except KeyError:
		return []
	return [media]

def imagehandler(ext, bin):
	if ext == ".json":
		data = json.loads(bin.decode("utf-8"))
		d = datetime.fromtimestamp(data["created_time"] / 1000)
		bin = f"{d:%Y-%m-%d %H:%M:%S}\n\n{data['text']}".encode("utf-8")
		ext = ".txt"
	return ext, bin


def get_episodes(html, url):
	eps = None
	next_page = None

	if "/user/" in url:
		data = re.search('<script id="__NEXT_DATA__" type="application/json">(.+?)</script>', html).group(1)
		data = json.loads(data)
		feeds = data["props"]["pageProps"]["initialState"]["api"]["feeds"]
		eps, next_page = get_episodes_from_posts(feed["post"] for feed in feeds.values())
		next_page_cache[url] = next_page

	else:
		data = json.loads(html)
		if "posts" not in data["data"]:
			raise LastPageError
		eps, next_page = get_episodes_from_posts(data["data"]["posts"])

	next_page_cache[url] = next_page
	eps.reverse()
	return eps

def get_medias(medias):
	for media in medias:
		if media["type"] == "VIDEO":
			if "resourceId" not in media:
				# the video is in boundContent
				continue
			yield f"https://obs.line-scdn.net/{media['resourceId']}/mp4"
		elif media["type"] == "PHOTO":
			yield f"https://obs.line-scdn.net/{media['resourceId']}/m800x1200"
		else:
			raise TypeError(f"Unknown media type: {media['type']}")

def get_next_page(html, url):
	u = next_page_cache.pop(url, None)
	if u:
		return u



#! python3

import re

from ..core import Episode

domain = ["tsundora.com"]
name = "tsundora"
noepfolder = True

def get_title(html, url):
	title = re.search(r'top_title">([^<]+)', html).group(1)
	return "[tsundora] " + title

def get_episodes(html, url):
	s = []
	# base = re.search("(https?://[^/]+)", url).group(1)
	match = None
	for match in re.finditer(
			r'href="(http://tsundora\.com/(\d+))"class="img_hover_trans"'
			'title="([^"]+)"', html):
		url, id, title = match.groups()
		s.append(Episode(id + " - " + title, url))
	return s[::-1]

def get_images(html, url):
	img = re.search(r'post-img">\s*<a href="([^"]+)"', html).group(1)
	return [img]

def get_next_page(html, url):
	match = re.search(r"rel='next' href='([^']+)'", html)
	if match:
		return match.group(1)
	


#! python3

import tkinter as tk
from tkinter import ttk

from .core import safe_tk

class Dialog:
	def __init__(self, parent, title="Dialog", on_closed=None):
		self.parent = parent
		self.result = None
		self.on_closed = on_closed
	
		# root
		self.root = tk.Toplevel(parent)
		self.root.title(safe_tk(title))
		
		# body
		self.body = ttk.Frame(self.root)
		self.create_body()
		self.body.pack()
		
		# button bar
		self.btn_bar = ttk.Frame(self.root)
		self.create_buttons()
		self.btn_bar.pack()
		
		# bind event
		self.root.bind("<Return>", self.resolve)
		self.root.bind("<Escape>", self.reject)
		
		# show dialog and disable parent
		self.root.grab_set()
		self.root.protocol("WM_DELETE_WINDOW", self.reject)
		self.root.focus_set()
		
	def create_body(self):
		ttk.Label(self.body, text="Default dialog").pack()
		
	def create_buttons(self):
		ttk.Button(
			self.btn_bar, text="確定", command=self.resolve, default="active"
		).pack(side="left")
		ttk.Button(
			self.btn_bar, text="取消", command=self.reject
		).pack(side="left")
		
	def resolve(self, _event=None):
		"""Apply and destroy dialog."""
		self.root.update_idletasks()
		self.result = self.apply()
		self.parent.focus_set()
		self.root.destroy()
		self.emit_closed()
		
	def reject(self, _event=None):
		"""Destroy dialog."""
		self.parent.focus_set()
		self.root.destroy()
		self.emit_closed()
		
	def emit_closed(self):
		if self.on_closed:
			self.on_closed(self.result)
		
	def apply(self):
		return True
		
	def wait(self):
		"""Wait the dialog to close."""
		self.root.wait_window(self.root)
		return self.result



#! python3

"""
http://www.177pic.info/html/2019/09/3114079.html
"""

import re
from html import unescape

from ..core import Episode

domain = ["www.177pic.info"]
name = "177"
noepfolder = True

def get_title(html, url):
	title = re.search(r'<h1[^>]*>([^<]+)', html).group(1)
	return unescape(title).strip()

def get_episodes(html, url):
	return [Episode("image", url)]

def get_images(html, url):
	imgs = []
	for match in re.finditer(r'data-lazy-src="([^"]+)', html):
		imgs.append(match.group(1))
	return imgs
	
def get_next_page(html, url):
	next_page = re.search(r'href="([^"]+)"><span><i class="be be-arrowright">', html)
	if next_page:
		return next_page.group(1)



from requests.utils import dict_from_cookiejar

from .grabber import grabhtml, grabimg

class ModuleGrabber:
	"""Bind grabber with module's header, cookie..."""
	def __init__(self, mod):
		self.mod = mod
		
	def html(self, url, **kwargs):
		return self.grab(grabhtml, url, **kwargs)
		
	def img(self, url, **kwargs):
		return self.grab(grabimg, url, **kwargs)
		
	def grab(self, grab_method, url=None, **kwargs):
		new_kwargs = {
			"header": self.get_header(),
			"cookie": self.get_cookie(),
			"done": self.handle_grab,
			"proxy": self.mod.config.get("proxy"),
			"verify": self.mod.config.getboolean("verify", True)
			}
		new_kwargs.update(kwargs)
		
		if hasattr(self.mod, "grabhandler"):
			result = self.mod.grabhandler(grab_method, url, **new_kwargs)
			if result:
				return result
		return grab_method(url, **new_kwargs)
		
	def get_header(self):
		"""Return downloader header."""
		return getattr(self.mod, "header", None)

	def get_cookie(self):
		"""Return downloader cookie."""
		cookie = getattr(self.mod, "cookie", {})
		config = getattr(self.mod, "config", {})
		
		for key, value in config.items():
			if key.startswith("cookie_"):
				name = key[7:]
				cookie[name] = value
				
		return cookie
		
	def handle_grab(self, session, _response):
		cookie = dict_from_cookiejar(session.cookies)
		config = getattr(self.mod, "config", None)
		if not config:
			return
			
		for key in config:
			if key.startswith("cookie_"):
				name = key[7:]
				if name in cookie:
					config[key] = cookie[name]



#! python3

"""this is gufeng module for comiccrawler

Example:

http://www.gufengmh.com/manhua/jueshiliandanshi/
https://www.gufengmh8.com/manhua/wodedabaojian/
"""

from html import unescape
from urllib.parse import urljoin
import re

from deno_vm import eval

from ..episode import Episode
from ..grabber import grabhtml

domain = ["www.gufengmh.com", "www.gufengmh8.com"]
name = "古風"

def get_title(html, url):
	return unescape(re.search("<h1><span>([^<]+)", html).group(1))

def get_episodes(html, url):
	id = re.search('manhua/([^/]+)', url).group(1)
	pattern = '<a href="(/manhua/{}/\d+.html)"[^>]+?>\s*<span>([^<]+)'.format(id)

	def create_ep(match):
		ep_url, title = [unescape(t) for t in match.groups()]
		return Episode(title, urljoin(url, ep_url))

	return [create_ep(m) for m in re.finditer(pattern, html)]

def get_images(html, url):
	js = re.search('(var siteName.+?)</script>', html, re.DOTALL).group(1)
	# http://www.gufengmh.com/js/config.js
	config = grabhtml(urljoin(url, "/js/config.js"))
	return eval("""
	const toastr = {
		options: {}
	};
	""" + config + js + """
	const domain = SinConf.resHost[0].domain[0];
	chapterImages.map(i => `${domain}/${chapterPath}${i}`);
	""")
	



#! python3

"""this is yandere module for comiccrawler

Ex:
	https://yande.re/post?tags=koikishi_purely_kiss

"""

import re
from html import unescape
from urllib.parse import urljoin

from ..core import Episode

domain = ["yande.re"]
name = "yande.re"
noepfolder = True

def get_title(html, url):
	title = re.search(r"<title>(.+?)</title>", html, flags=re.DOTALL).group(1)
	title = title.strip("/")
	title = title.replace(" | yande.re", "")
	return "[yande.re] " + title
	
def get_episodes(html, url):
	s = []
	base = re.search("(https?://[^/]+)", url).group(1)
	for match in re.finditer(r'href="(/post/show/(\d+)[^"]*)"', html):
		u = match.group(1)
		title = match.group(2)
		e = Episode(title, base + u)
		s.append(e)
	return s[::-1]
			
def get_images(html, url):
	
	# Original
	img = re.search(
		r'class="original-file-unchanged"[^>]*?href="([^"]+)"',
		html
	)
	
	# Larger version
	if not img:
		img = re.search(r'id="highres" href="([^"]+)"', html)
		
	img = img.group(1)

	return [img]

def get_next_page(html, url):
	u = re.search(r'rel="next" href="([^"]+)"', html)
	if u:
		return urljoin(url, unescape(u.group(1)))
		


#! python3

import json
import re
from urllib.parse import urlparse, parse_qs

from ..episode import Episode
from ..error import SkipPageError
from ..url import update_qs
from ..grabber import grabhtml

domain = ["beta.sankakucomplex.com"]
name = "Sankaku Beta"
noepfolder = True

class ExpireError(Exception):
	pass

def get_query(url, name):
	query = urlparse(url).query
	return parse_qs(query)[name][0]

def get_title(html, url):
	return "[sankaku] {}".format(get_query(url, "tags"))

next_page_cache = {}

def get_episodes(html, url):
	if re.match(r"https://beta\.sankakucomplex\.com/\?", url):
		next_page_cache[url] = update_qs("https://capi-v2.sankakucomplex.com/posts/keyset?lang=en&default_threshold=1&hide_posts_in_books=never&limit=40", {
			"tags": get_query(url, "tags")
		})
		raise SkipPageError
		
	data = json.loads(html)
	next = data["meta"]["next"]
	# data_len_cache[url] = len(data)
	eps = [
		Episode(
			str(e["id"]),
			"https://beta.sankakucomplex.com/post/show/{}".format(e["id"]),
			image=e["file_url"]
		) for e in data["data"]
	]
	
	if next:
		next_page_cache[url] = update_qs(url, {
			"next": next
		})
	
	return eps[::-1]

def get_images(html, url):
	id = re.search("post/show/(\d+)", url).group(1)
	data = grabhtml("https://capi-v2.sankakucomplex.com/posts?lang=english&page=1&limit=1&tags=id_range:{}".format(id))
	data = json.loads(data)
	return data[0]["file_url"]
	
def get_next_page(html, url):
	if url in next_page_cache:
		return next_page_cache.pop(url)
			
def redirecthandler(response, crawler):
	if re.match(r"https://chan\.sankakucomplex\.com/.+\.(png|jpg)", response.url):
		raise ExpireError

def errorhandler(err, crawler):
	if isinstance(err, ExpireError):
		crawler.ep.image = None
		crawler.html = None
		